;;;; File: am-services.lsp
;;; SCCS Version: %W%
;;; Contains: General code for the AM.
;;; Author: Jeff Dalton and Richard Kirby (rbk)
;;; Created: Fri Nov  2 13:37:42 1990
;;; Updated: Sun Jul  9 19:05:06 1995 by Jeff Dalton
;;; Release Version: %Y%
;;; Copyright: (c) 1992, 1993, 1994, AIAI, University of Edinburgh
;;; This material may be reproduced by or for the U.S. Government pursuant
;;; to the copyright license under the clause at DFARS 252.227-7032
;;; (June 1975) -- Rights in Technical Data and Computer Software (Foreign).

(in-package :oplan)

;; :SINGLE-STEP handler, which presents a menu to the user for interrogating
;; the agenda tables, and for subverting the agenda orderings.

(defun am-single-step-init (value)
  ;; Called by mcc-handler to set single-step mode.
  (setq *am-single-step* value))

(defun am-single-step-handler ()
  (loop
    (ecase (am-single-step-menu-choice)
      (:triggered
       (do-display-agenda-table *debug-io*)
       (terpri *debug-io*))
      (:untriggered
       (do-display-untriggered-agenda-table *debug-io*)
       (terpri *debug-io*))
      (:alt
       (do-display-alt-agenda-table *debug-io*)
       (terpri *debug-io*))
      (:top
       (return))
      (:any
       (when (do-process-any-agenda-entry *debug-io*)
	 (return)))
      (:user
       (let ((ae (make-agenda-rec :trigger t :body '(:user :user-request))))
	 (assign-event-defaults ae)
	 (add-to-agent-agenda ae)))
      (:break
       (break "Return to continue with AM Interrogation."))
      (:quit
       (setq *am-single-step* nil)
       (return)))))

(defun am-single-step-menu-choice ()
  (menu-request
    `("-heading"
      ,(format nil "AM Single Step in cycle ~A" *cycle-count*)
      "Display the Agenda Table=:triggered"
      "Display the Untriggered Agenda Table=:untriggered"
      "Display the Alternatives Agenda Table=:alt"
      "Process the top agenda entry=:top"
      "Process any agenda entry=:any"
      "Schedule KS-USER=:user"
      "BREAK IN=:break"
      "-line"
      "Quit single step mode=:quit")))

(defun do-display-agenda-table (stream)
  (do-display-agent-agenda-table stream)
  (do-display-triggered-agenda-table stream))

(defun do-display-agent-agenda-table (stream)
  (when (agent-agenda-p)
    ;; /\/: Priorities?
    (format stream "~%The Agent Agenda:~%")
    (let ((*print-case* :downcase)
	  (*print-pretty* t)
	  (*print-level* 4)
	  (*print-length* 4))
      (dolist (ae (agent-agenda-entries))
	(do-display-agenda-entry ae stream)))))

(defun do-display-triggered-agenda-table (stream)
  (format stream "~%The Agenda Table:~%")
  (let ((*print-case* :downcase)
	(*print-pretty* t)
	(*print-level* 4)
	(*print-length* 4))
    (dolist (ae *agenda-table*)
      (do-display-agenda-entry ae stream))))

(defun do-display-untriggered-agenda-table (stream)
  (format stream "~%The Untriggered Agenda Table:~%")
  (let ((*print-case* :downcase)
	(*print-pretty* t)
	(*print-level* 4)
	(*print-length* 4))
    (dolist (ae (db-request :REQUEST-UNTRIGGERED-AGENDA-TABLE))
      ;; /\/: What about priorities?  OK as-is?
      (do-display-agenda-entry ae stream))))

(defun do-display-agenda-entry (ae stream)
  (mon-format stream
      "~&~:@(~A~)/~A: ~
       ~:I~:_Priority ~A, ~:_Level ~A, ~:_Branch-1 ~A, ~:_Branch-n ~A, ~
          ~:_Cost ~A, ~:_Trigger ~A: ~:_~A, ~:_Info ~A~%"
      (ag-id ae)
      (ag-stage ae)
      (ag-priority ae)
      (ag-level ae)
      (ag-branch-1 ae)
      (ag-branch-n ae)
      (ag-cost ae)
      (ag-trigger ae)
      (ag-body ae)
      (ag-info ae)))

(defun do-process-any-agenda-entry (io)
  ;; /\/: We lack a mechanism for putting a plan ae ahead of the agent
  ;; agenda, or even to take a particular agent ae.
  (when (agent-agenda-p)
    (tell-am-user io "Can't when the agent agenda is not empty.")
    (return-from do-process-any-agenda-entry nil))
  (let* ((n (ask-am-user io "AE-"))
	 (id (concat-name "AE-" n))
	 (ae (find id *agenda-table* :key #'ag-id)))
    (cond (ae
	   (setq *agenda-table*
		 (cons ae
		       (remove-1-eq ae *agenda-table*)))
	   t)
	  (t
	   (tell-am-user io "No ~A" id)
	   nil))))
    
(defun ask-am-user (io question &rest format-args)
  ;; /\/ There's a similar routine with the same name in the TA.
  ;; /\/ And others, in other places.
  (format io "~&~?" question format-args)
  (read-line io))

(defun tell-am-user (io message &rest format-args)
  (format io "~&~?~%" message format-args))


;;; Agenda priorities

(defparameter *agenda-priorities*
  '((:KILLJOY      . 10000)
    (:KILL         . 1000) 		; new for 2.0 /\/
    (:INIT         . 100)
    (:NEWDOMAIN    . 90)
    (:SET-TASK     . 80)
    (:USER         . 70)
    (:EXTRACT      . 65)
    (:POISON-STATE . 60)
    (:OR           . 55)
    (:EXPAND       . 50)
    (:ACHIEVE      . 40)
    (:CONDITION    . 30)
    (:COMPUTE      . 30)		;same as :CONDITION
    (:BIND         . 20)))

(defparameter *default-agenda-priority* 10)

(defun-inline get-agenda-priority (agenda)
  (or (cdr (assoc (car (ag-body agenda)) *agenda-priorities*))
      *default-agenda-priority*))

#| Future note:
For CONDITION agenda entries, probably want to rate unsupervised tieups higher
than only_use_for_query tieups.
|#

(proclaim '(notinline assign-agenda-priority))

(defun assign-agenda-priority (ag)
  (let ((priority (get-agenda-priority ag)))
    (if (ag-priority ag)
	(assert (= (ag-priority ag) priority))
      (setf (ag-priority ag)
	    priority))))

(defun assign-agenda-id (ag)
  (unless (ag-id ag)
    (setf (ag-id ag) (pop-gensym "AE-"))))

(defun assign-agenda-defaults (ag)
  (macrolet ((default ((accessor item) value)
	       `(unless (,accessor ,item)
		  (setf (,accessor ,item) ,value))))
    (assert (null (ag-id ag)))
    (assert (null (ag-priority ag)))
    (assign-agenda-id ag)
    (assign-agenda-priority ag)
    (case (ag-type ag)
      (:poison-state
       (default (ag-trigger  ag) t)
       (default (ag-branch-1 ag) 1)
       (default (ag-branch-n ag) 1)
       (default (ag-level    ag) -1))
      (t
       (default (ag-trigger  ag) t)
       (default (ag-branch-1 ag) :inf)	;/\/
       (default (ag-branch-n ag) :inf)	;/\/
       (default (ag-level    ag) (&+ 0 (current-agenda-level))))) ;was 1+ /\/
    ag))

(defun assign-event-defaults (ag)
  ;; /\/: id and priority not always null because KS-USER can send its
  ;;      agenda-entry back to the AM.
  ;(assert (null (ag-id ag)))
  ;(assert (null (ag-priority ag)))
  (assign-agenda-id ag)
  (assign-agenda-priority ag)
  (macrolet ((default ((accessor item) value)
	       `(unless (,accessor ,item)
		  (setf (,accessor ,item) ,value))))
    (default (ag-trigger  ag) t)
    (default (ag-branch-1 ag) 1)
    (default (ag-branch-n ag) 1)
    (default (ag-level    ag) -1)
    ag))

(defun assign-alt-agenda-defaults (ag)
  (assign-agenda-id ag)
  (assign-agenda-priority ag)
  (macrolet ((default ((accessor item) value)
	       `(unless (,accessor ,item)
		  (setf (,accessor ,item) ,value))))
    ;(default (ag-trigger  ag) t)
    (default (ag-branch-1 ag) :inf)	;/\/
    (default (ag-branch-n ag) :inf)	;/\/
    (default (ag-level    ag) (&+ 0 (current-agenda-level))) ;was 1+ /\/
    ag))

(defun check-agenda-entry (ag)
  (assert (and (symbolp (ag-id ag)) (not (null (ag-id ag)))))
  (assert (not (null (ag-trigger ag))))
  (assert (numberp (ag-priority ag)))
  (assert (and (consp (ag-body ag)) (symbolp (car (ag-body ag)))))
  (assert (or (numberp (ag-level ag))
	      (and (member (ag-type ag) '(:or :bind :user))
		   (infp (ag-level ag)))))
  (assert (&numberp (ag-branch-1 ag)))
  (assert (&numberp (ag-branch-n ag)))
  ag)

(defun current-agenda-level ()
  (let ((ae *kp-processing-which-ae*))
    (if ae
	(or (ag-level ae) -1)
      -1)))
  
#+:lose
(defun calculate-agenda-entries-priorities ()
  (mapc #'assign-priority *agenda-table*)
  (setq *agenda-table*
	(stable-sort *agenda-table*
		     #'(lambda (x y)
			 ;; /\/: Not a <: 2nd compare s.b. only if = priority
			 (or (> (ag-priority x) (ag-priority y))
			     (< (read-from-string (symbol-name (ag-id x))
						  nil nil :start 3)
				(read-from-string (symbol-name (ag-id y))
						  nil nil :start 3)))))))

;;; /\/: If agenda entries have uniquely numbered ids, stable-sort
;;; shouldn't be necessary.  Also, we shouldn't have to call read-
;;; from-string each time!  So...

(defun request-current-agenda-table ()
  (ipc-with-own-copy (table (db-request :REQUEST-CURRENT-AGENDA-TABLE)
			    (copy-list table))
    (mapc #'check-agenda-entry table)
    (setq *agenda-table*
	  (sort table #'ag>))))

#+:undef
(defun ag> (ag1 ag2)
  (or (> (ag-priority ag1) (ag-priority ag2))
      (and (= (ag-priority ag1) (ag-priority ag2))
	   (< (ag-id-number ag1) (ag-id-number ag2)))))

(defun ag> (a b)			;is a better than b?
  (cond
	((fix> (ag-priority a) (ag-priority b))
	 t)
	((fix> (ag-priority b) (ag-priority a))
	 nil)
	(t
	 (fix< (ag-id-number a) (ag-id-number b)))))

(defun ag-id-number (ag)
  (or (get (ag-id ag) 'ag-id-number)
      (setf (get (ag-id ag) 'ag-id-number)
	    (read-from-string (symbol-name (ag-id ag))
			      nil nil :start 3))))


;;;; Alternative management.

(defun post-alternative (alt)
  (db-request :ADD-ALT-ENTRY alt))

(defun create-alt-record (alt-agenda)
  (make-alternative-rec :id (pop-gensym "ALT-")
			:alt-agenda alt-agenda
			:processing (ag-id *kp-processing-which-ae*)))


;; Very simple, just pops from alt agenda table.  The entries are
;; already sorted, so that the best are first, in the DM.
#+:undef
(defun choose-alt (reason alts)
  (declare (ignore reason))
  (pop alts))

;;; /\/: It may work to take all :BIND and :OR alts and take the best
;;; from them.  Most recent is suspect in :OR since it may not be the
;;; alt that contains other branches from the tree but instead the
;;; alt that reachieves "any old how".  Indeed, it may be a problem
;;; in general that that alt is more recent?

;;; /\/: A problem with going depth-first for, say, :COULD-NOT-BIND,
;;; thinking it will only invlove BINDs, is that the poison may be from
;;; an OR.

(defun choose-alt (reason alts)
  (ecase (car reason)
    ((:agenda-slice-exhausted)
     ;; Take best-rated alt.  Forget about depth-firstiness.
     (car alts))
    ((:poison-state)
     (case (cadr reason)
       ((:COULD-NOT-FIND-BINDING
	 :COULD-NOT-BIND
	 :COULD-NOT-RESTRICT
	 :COULD-NOT-ADD-LINK
	 )
	;; Consider most recent alt
	(consider-most-recent-alts reason alts))
       (t
	;; Take best-rated alt
	(car alts))))))

(defun consider-most-recent-alts (reason alts)
  (declare (ignore reason))
  (let ((most-recent-first
	 (sort (copy-list alts)
	       #'(lambda (a b)
		   (> (alt-id-number a)
		      (alt-id-number b)))))
	(best-rated
	 (car alts)))
    (dolist (a most-recent-first
	       best-rated)		;default
      (let ((alt-ae (alt-alt-agenda a)))
	(when (null alt-ae)
	  (return best-rated))
	(case (ag-type alt-ae)
	  ((:bind)
	   (dev-warn "Taking most recent ~S" (ag-type alt-ae))
	   (return a))
	  ((:or)
	   (if (or-is-reachieve-p alt-ae)
	       (dev-warn "Skipping reachieve")
	     (progn
	       (dev-warn "Taking most recent ~S" (ag-type alt-ae))
	       (return a))))
	  (t
	   (return best-rated)))))))

(defun or-is-reachieve-p (ae)
  ;; See set-up-alternative-with-opposite in KS-OR
  (assert (eq :or (car (ag-body ae))))
  (destructuring-bind (cond contribs satisfy-p &rest or-tree)
                      (cdr (ag-body ae))
    (declare (ignore cond contribs satisfy-p))
    (let* ((choice (car or-tree))
	   (or-rec (car choice))
	   (actions (or-info-rec-actions or-rec))
	   (action-1 (car actions)))
      (eq (car action-1) :reachieve))))	

(defun alt-id-number (alt)
  (or (get (alt-id alt) 'alt-id-number)
      (setf (get (alt-id alt) 'alt-id-number)
	    (read-from-string
	      (symbol-name (alt-id alt)) nil nil :start 4))))

(defun pick-an-alternative (reason)
  ;; The reason is typically (:POISON-STATE keyword object...).
  (incf *n-alts-chosen*)
  (when (eq (car reason) :poison-state)
    (incf *n-poisons*))
  (let* ((alts (db-request :REQUEST-ALTS-AGENDA-TABLE))
	 (alt (choose-alt reason alts)))
    (setq *slice-remaining* *agenda-slice*) 	;fresh start
    ;; Go back to the context that had the alternative.
    (db-request :SET-CONTEXT (alt-context alt))
    ;; Check consistency
    (when (db-request :CONTEXT-WAS-POISONED)
      (cerror "Continue anyway"
	      "Alt in poisoned context.~%Poisoned because ~A~%Alt: ~S"
	      (db-request :CONTEXT-WAS-POISONED)
	      alt))
    ;; Immediately push to a new context to preserve the alternative context.
    (db-request :PUSH-CONTEXT)
    ;; We have to combine the deletion of the agenda entry that was
    ;; being processed with the addition of the alternative agenda
    ;; entry to avoid triggering :planner-finished by mistake, which
    ;; would happen if the entry being deleted was the last entry in
    ;; the context.  Note that we can't simply add before deleting,
    ;; because the two entries might have the same id and the wrong
    ;; one (or both!) might be deleted.
    (if (alt-alt-agenda alt)
	(db-request :EXCHANGE-AGENDA-ENTRY
		    (alt-processing alt)
		    (alt-alt-agenda alt))
      ;; This is an alt created by switch-alternatives, so both the
      ;; alt-agenda and processing slots are nil.
      (assert (null (alt-processing alt))))
    #|
    ;; Remove the agenda entry that was being processed.
    (db-request :PROCESSED-AGENDA-ENTRY (alt-processing alt))
    ;; Add in the alternative agenda entry.
    (db-request :ADD-AGENDA-ENTRY (alt-alt-agenda alt))
    |#
    (unless (eq (db-request :PROCESSING-ALT-AGENDA (alt-id alt)) :OK)
      (error "DM rejected alternative ~S." alt))
    (request-current-agenda-table)))

;;; Switching alternatives

;;; /\/: When the DM adds an alt, it records the current context in the
;;; alt and pushes a context.  This preserves the alt's context.  When
;;; pick-an-alternative picks an alt, it also pushes a context.  Here we
;;; add an alt then immediately pick one (typically a different one).
;;; So the new context created when we add the alt is never used.

(defun maybe-switch-alternatives ()
  (if (db-request :funcall 'atm-exists-better-alternative-p)
      (switch-alternatives)
    (progn
      ;; Since we'd just return to the current state anyway, we'll
      ;; just give it a fresh slice w/o going via alternatives.
      (dev-debug :user-request "No better alternative")
      (dev-warn "No better alternative")
      (setq *slice-remaining* *agenda-slice*))))

(defun switch-alternatives ()
  (dev-debug :user-request "Switching alternatives")
  (dev-warn "Switching alternatives")
  (let ((alt (make-alternative-rec
	       :id (pop-gensym "ALT-")
	       :alt-agenda nil
	       :processing nil)))
    (post-alternative alt)		;to remember the current context
    (pick-an-alternative '(:agenda-slice-exhausted))))

;;; Alt agenda display

(defun do-display-alt-agenda-table (stream)
  (format stream "~%The Alternatives Agenda Table:~%")
  (let ((*print-case* :downcase)
	(*print-pretty* t)
	(*print-level* 4)
	(*print-length* 4))
    (dolist (alt (db-request :REQUEST-ALTS-AGENDA-TABLE))
      (let ((ae (alt-alt-agenda alt)))
	(mon-format stream
	    "~A: ~:IContext: ~A Processing ~A ~:_Agenda: ~A/~A: ~:_~A~%"
	    (string (alt-id alt))		;to get upper case
	    (alt-context alt)
	    (string (alt-processing alt)) 	;to get upper case
	    (string (ag-id ae))			;to get upper case
	    (ag-stage ae)
	    (ag-body ae))))))

;;; ---------------------------- Change History ----------------------------
;;; (Who)   (When) 		     (What)
;;;

