;;;; File: am-toplevel.lsp
;;; SCCS Version: %W%
;;; Contains: Top-level processing for the AM.
;;; Author: Jeff Dalton and Richard Kirby
;;; Created: Sun Apr 15 20:21:56 1990
;;; Updated: Wed Nov  9 19:31:31 1994 by Jeff Dalton
;;; Release Version: %Y%
;;; Copyright: (c) 1992, 1993, 1994, AIAI, University of Edinburgh
;;; This material may be reproduced by or for the U.S. Government pursuant
;;; to the copyright license under the clause at DFARS 252.227-7032
;;; (June 1975) -- Rights in Technical Data and Computer Software (Foreign).

;;; /\/: There should be a mechanism for keeping track of statistics,
;;; clearing them, etc.

(in-package :oplan)

(defvar *am-single-step* nil)	;true for single-step mode

(defvar *agent-agenda* nil)	;owned by AM; not context-layered
(defvar *agenda-table* nil)	;copy of plan agenda owned by DM

(defvar *cycle-count* 0)	;number of AEs processed; never reset

(defvar *cycle-count-base*	;for the elapsed-cycles statatistic
  *cycle-count*)		;  reset on init or replan

(defparameter *agenda-slice* nil)
(defvar *slice-remaining* *agenda-slice*)

(defvar *n-alts-chosen* 0)	;statistic
(defvar *n-poisons* 0)		;statistic

(defvar *kp-processing-which-ae* nil)
(defvar *ready-kps* nil)


;;;; Startup and event-handler

(defun am-startup ()

  ;; This is the initial function for the AM process.  After startup,
  ;; it is not called again.  Instead, am-event-handler is called
  ;; whenever a message arrives.

  (unless (is-init *whoami*)
    (dev-error *whoami* "is-init failed."))
  (setf (component-init-fn *pprocess*) #'am-init)
  (setf (component-single-step-fn *pprocess*) #'am-single-step-init)
  ;; Make breaks, etc. go to the debug window.
  (x-redirect-errors :amout)
  (clear-screen *debug-io*)
  ;; Register with IM.
  (dev-debug :information "Connecting to IM via MCC.")
  (mcc-register-with-im)
  (ipc-set-event-handler 'am-event-handler))

(defun am-init ()
  ;; Called when the AM receives an :INIT message on :MCC
  (clear-screen *debug-io*)
  (clear-pop-gensymtable)
  (setq *agenda-table* nil)
  ;; /\/: Should we clear the agent agenda?
  (setq *slice-remaining* *agenda-slice*)
  (setq *cycle-count-base* *cycle-count*
	*n-alts-chosen* 0
	*n-poisons* 0)
  (dev-debug :information "Initialised"))

(defun am-event-handler (self)
  (let ((e (next-event self)))
    (case (message-route e)
      (:mcc (mcc-handler (message-contents e)))
      (otherwise
       (ipc-handle-message self e))))
  (send-any-triggered-agenda-entries-off-to-waiting-kps))


;;;; The agent agenda

;;; Kept as a sorted list, higher-priority entries first.

(defun-inline agent-agenda-p ()
  *agent-agenda*)

(defun agent-agenda-entries ()
  *agent-agenda*)

(defun add-to-agent-agenda (ae)
  (check-agenda-entry ae)
  (setq *agent-agenda*
	(merge 'list *agent-agenda* (list ae) #'> :key #'ag-priority)))

(defun-inline first-from-agent-agenda ()
  (car *agent-agenda*))

(defun-inline take-from-agent-agenda ()
  (pop *agent-agenda*))


;;;; The plan agenda

(defun-inline plan-agenda-p ()
  *agenda-table*)

(defun-inline first-from-plan-agenda ()
  (car *agenda-table*))

(defun-inline take-from-plan-agenda ()
  (when *slice-remaining*
    (decf *slice-remaining*))
  (pop *agenda-table*))


;;;; Message handlers

;;; Via :agendain

;;; /\/: Need to do something about these ack (:ok) messages.

(defmessage ((:am :agendain) :agenda) (agenda)
  (assign-agenda-defaults agenda)
  (db-request :ADD-AGENDA-ENTRY agenda))

(defmessage ((:am :agendain) :event) (agenda)
  (assign-event-defaults agenda)
  (add-to-agent-agenda agenda))

(defmessage ((:am :agendain) :alt-agenda) (alt-agenda)
  (assign-alt-agenda-defaults alt-agenda)
  (post-alternative (create-alt-record alt-agenda))
  #+:undef (ipc-write stream :OK)
  )

(defmessage ((:am :agendain) :pick-alt) (reason)
  (dev-debug :warning "Asked to pick alt because ~S" reason)
  (pick-an-alternative reason)
  #+:undef (ipc-write stream :OK)
  )

(defmessage ((:am :agendain) :replanning) ()
  (setq *cycle-count-base* *cycle-count*
	*n-alts-chosen* 0
	*n-poisons* 0))

(defmessage ((:am :agendain) :elapsed-cycles) ()
  (- *cycle-count* *cycle-count-base*))

(defmessage ((:am :agendain) :get-plan-statistics) ()
  `((:am-cycles         .  ,(- *cycle-count* *cycle-count-base*))
    (:n-alts-chosen     .  ,*n-alts-chosen*)
    (:n-poisons         .  ,*n-poisons*)
    (:n-alts-remaining  .  ,(n-alts-remaining)) ))

(defun n-alts-remaining ()
  (db-request :funcall 'atm-get-number-of-alternatives))

;;; Via :triggersin

(defmessage ((:am :triggersin) :invalidate-agenda-table) ()
  (request-current-agenda-table))

;;; Via :kpready

;;; /\/: Storing :kpready is bogus.  It's supposed to be a channel
;;; direct to a particular KP.  In the single-process system we can
;;; just call send-to-client giving the KP as the recipient.  In
;;; other systems, some kind of association between client ids
;;; and communication channels has to be kept.

(defmessage ((:am :kpready) :kpready) (kp)
  (store-ready-kp :kpready kp))

;;; Handle :KPREADY status

(defun store-ready-kp (stream ident)
  (when *kp-processing-which-ae*
    (db-request :PROCESSED-AGENDA-ENTRY (ag-id *kp-processing-which-ae*))
    (request-current-agenda-table)
    (setq *kp-processing-which-ae* nil))
  (push (cons stream ident) *ready-kps*))

(defun kp-ready-p ()
  *ready-kps*)


;;;; Sending to knowledge source platforms

;;; /\/: Some things give the impression that we can handle multiple
;;;      KPs, but in fact lots of things won't work if we try.

;;; /\/: When the agenda entry is taken from the _agent_ agenda, we still
;;;      tell the DM to remove it from the triggered aganda table, even
;;;      though it isn't in the table.  The DM can handle this, but it
;;;      still shouldn't happen.  Consider a different type of agenda
;;;      entry (type as in typep) or a record of which agenda it came
;;;      from.

;;; /\/: send-any-triggered-agenda-entries-off-to-waiting-kps should be
;;;      part of the event handler, not a separate fn.

;;; We can't swich alts when the KP is in the middle of processing an
;;; agenda entry, because the KS miight send the AM some messages that
;;; make sense only in the current alt.  (E.g. it might ask the AM to
;;; add an agenda entry.)  So we consider switching only when (kp-ready-p)
;;; is true.  Since there's only one KP this is sufficient.

(defun send-any-triggered-agenda-entries-off-to-waiting-kps ()
  (when (kp-ready-p)
    (when (and *slice-remaining*
	       (<= *slice-remaining* 0)
	       ;; Don't switch if a poison ae will be coming along.
	       (not (db-request :context-was-poisoned)))
      (maybe-switch-alternatives))
    (when (or (plan-agenda-p)
	      (agent-agenda-p))
      (send-triggered-agenda-entry-to-waiting-kp)
      (when (kp-ready-p)
	(error "Can't handle > 1 ready KPs.")))))

(defun send-triggered-agenda-entry-to-waiting-kp ()
  (incf *cycle-count*)
  (when *am-single-step*
    (am-single-step-handler))
  (let ((a-kp (pop *ready-kps*))
	(ae (get-agenda-entry-to-process)))
    ;; a-kp is something like (:KPREADY . :KP-1)
    (dev-debug :user-request "Cycle ~S: Processing ~S/~S,~(~A~)"
	       *cycle-count* (ag-id ae) (ag-stage ae) (ag-level ae))
    (dev-debug :information "Sending a triggered agenda onto KP ~A~%"
	       (cdr a-kp))
    (setq *kp-processing-which-ae* ae)
    (ipc-write-client (car a-kp) (cdr a-kp)
       (mcc-make-message :KPREADY :AGENDA ae))))


;;;; Picking an agenda entry to process

(defun get-agenda-entry-to-process ()
  (cond ((not (plan-agenda-p))
	 ;; Nothing on the plan agenda
	 (if (agent-agenda-p)
	     ;; But something on the agent agenda
	     (take-from-agent-agenda)
	   ;; Neither plan nor agent agenda has anything for us to do.
	   ;; This isn't supposed to happen.
	   (error "Nothing to do.")))
	;; Something on the plan agenda ...
	((agent-agenda-p)
	 ;; ... and on the agent agenda -- see which has higher priority;
	 ;; prefer agent agenda if there's a tie.
	 (let ((plan-entry (first-from-plan-agenda))
	       (agent-entry (first-from-agent-agenda)))
	   (if (>= (ag-priority agent-entry) (ag-priority plan-entry))
	       (take-from-agent-agenda)
	       (take-from-plan-agenda))))
	(t
	 ;; ... but not on the agent agenda -- take plan agenda entry.
	 (take-from-plan-agenda))))

;;; ---------------------------- Change History ----------------------------
;;; (Who)   (When) 		     (What)
;;;
