;;;; File: atm.lsp
;;; SCCS Version: %W%
;;; Contains: Agenda Table Manager.
;;; Author: Richard Kirby (rbk) and Jeff Dalton
;;; Created: Wed Nov 27 18:38:45 1991
;;; Updated: Sun Jul  9 18:07:12 1995 by Jeff Dalton
;;; Release Version: %Y%
;;; Copyright: (c) 1992, 1993, 1994, AIAI, University of Edinburgh
;;; This material may be reproduced by or for the U.S. Government pursuant
;;; to the copyright license under the clause at DFARS 252.227-7032
;;; (June 1975) -- Rights in Technical Data and Computer Software (Foreign).

;;; Manages the Agenda Table and the Alternatives Table.

(in-package :oplan)

;;; Functions that can be redefined by the user:

(proclaim '(notinline atm-agenda-cost-fn
	              atm-agenda-entry-cost-fn
	              atm-state-rating-fn
	              atm-alt-rating-fn))

;;; Triggered and untriggered agenda tables (context-layered).

(defvar *triggered-agenda-table* nil)
(defvar *untriggered-agenda-table* nil)

;;; Agenda cost estimates

;;; *Agenda-cost* (context-layered) is a estimated, collective cost of
;;; all entries in both the triggered and untriggered agenda tables.
;;; It is used when rating alternatives.

(defvar *agenda-cost* 0)

;;; atm-agenda-entry-cost-fn estimates the cost of a single agenda
;;; entry.  It can return a positive integer or :inf.

#+:undef
(defun atm-agenda-entry-cost-fn (ag)
  (case (ag-type ag)
    (:expand
     1)
    (:achieve
     1)
    (t
     0)))

(defun atm-agenda-entry-cost-fn (ag)
  (case (ag-type ag)
    (:expand
     (round 4 (fix-max (ag-level ag) 1)))
    (:achieve
     (round 4 (fix-max (ag-level ag) 1)))
    (t
     1)))

;;; atm-agenda-cost-fn estimeates the collective cost of the triggered
;;; and untriggered agenda tables.

(defun atm-agenda-cost-fn ()
  (round (ctxt-symbol-value '*agenda-cost*) 2))


;;;; Agenda table operations.

;;; atm-add-untriggered-agenda-entry -- called from message handling
;;; loop when received a :ADD-AGENDA-ENTRY message from the AM.

(defun atm-add-untriggered-agenda-entry (ag)
  (&incf (ctxt-symbol-value '*agenda-cost*)
	 (setf (ag-cost ag)
	       (atm-agenda-entry-cost-fn ag)))
  (push ag (ctxt-symbol-value '*untriggered-agenda-table*)))

;;; atm-remove-untriggered-agenda-entry -- called only from triggering
;;; code.  N.B. arg must be the same (eq) ag, not a copy.

(defun atm-remove-untriggered-agenda-entry (ag)
  (setf (ctxt-symbol-value '*untriggered-agenda-table*)
	(remove-1-eq ag (ctxt-symbol-value '*untriggered-agenda-table*))))

;;; atm-add-triggered-agenda-entry -- called from triggering code.

(defun atm-add-triggered-agenda-entry (ag)
  (push ag (ctxt-symbol-value '*triggered-agenda-table*)))

;;; atm-remove-triggered-agenda-entry -- called from message handling
;;; loop when received a :PROCESSED-AGENDA-ENTRY message from the AM.

(defun atm-remove-triggered-agenda-entry (ag-id)
  (let ((ag (find ag-id
		  (ctxt-symbol-value '*triggered-agenda-table*)
		  :key #'ag-id)))
    (when ag
      (&decf (ctxt-symbol-value '*agenda-cost*)
	     (ag-cost ag))
      (setf (ctxt-symbol-value '*triggered-agenda-table*)
	    (remove-1-eq ag (ctxt-symbol-value '*triggered-agenda-table*)))
      ag-id)))

;;; atm-get-untriggered-agenda-table -- called from triggering code.

(defun atm-get-untriggered-agenda-table ()
  (ctxt-symbol-value '*untriggered-agenda-table*))

;;; atm-get-triggered-agenda-table -- called from message handling
;;; loop when received a :REQUEST-CURRENT-AGENDA-TABLE message from
;;; the AM.

(defun atm-get-triggered-agenda-table ()
  (ctxt-symbol-value '*triggered-agenda-table*))

;;; atm-empty-tables-p -- called from triggering code to see if the
;;; agenda-tables are now empty.

(defun atm-empty-tables-p ()
  (let ((triggered (ctxt-symbol-value '*triggered-agenda-table*))
	(untriggered (ctxt-symbol-value '*untriggered-agenda-table*)))
    (and (null triggered)
	 ;; The following is equivalent to: (= (length untriggered) 1)
	 (and untriggered (null (cdr untriggered))))))

;;; atm-clear-agenda-tables -- called only by dm-init.

(defun atm-clear-agenda-tables ()
  ;; N.B. wipes out context record and everything.
  (setq *triggered-agenda-table* nil
	*untriggered-agenda-table* nil
	*agenda-cost* 0))


;;;; The alternatives.

;;; N.B. The alt table is not context-layered.

(defvar *alternatives-table* nil)

;;; Alt ratings are used to select the best alternative when a choice
;;; is made.  Lower values are considered better.  That is, the values
;;; represent a cost.

;;; atm-alt-rating-fn estimates the total cost of the best path through
;;; the current state to a solution (as in the A* algorithm).

(defun atm-alt-rating-fn (alt)
  (if (eq (car (ag-body (alt-alt-agenda alt)))
	  :NO-MORE-ALTERNATIVES)
      ;; Make sure :no-more-alternatives alts are chosen last.
      most-positive-fixnum
    ;; All other alts get a computed rating.
    (&+ (atm-state-rating-fn)		;cost to get here
	(atm-agenda-cost-fn))))		;cost from here to a solution

;;; atm-state-rating-fn returns the cost already paid to reach the
;;; current state.

(defun atm-state-rating-fn ()
  ;; Keep simple at moment - just return the number of nodes in the current
  ;; plan state.  NOTE LOW IS BEST!
  (let ((count 0))
    (walk-nodes
      #'(lambda (n)
	  (declare (ignore n))
	  (incf count)))
    (assert (= (* 2 count) (node-end-count)))
    (assert (> count 0))
    (if (> count 0)
	count
      most-positive-fixnum)))

;;; atm-exists-better-alternative-p rates the current state and
;;; compares it to the best available alternative.

(defparameter *switch-penalty* 0)

(defun atm-exists-better-alternative-p ()
  (and *alternatives-table*
       (&< (&+ (alt-rating (car *alternatives-table*))
	       *switch-penalty*)
	   (atm-alt-rating-fn :current-state))))

;;; atm-add-alternative -- called from message handling loop when
;;; received a :ADD-ALT-ENTRY message from the AM.

(defun atm-add-alternative (alt)
  (let ((context (contextnum *cucontext*)))
    (setf (alt-context alt) context)
    (setf (alt-rating alt) (atm-alt-rating-fn alt))
    (push alt *alternatives-table*)
    (setq *alternatives-table*
	  (stable-sort *alternatives-table*
		       #'&<
		       :key #'alt-rating))
    (pushcontext)
    t))

;;; atm-remove-alternative -- called from message handling loop when
;;; received a :PROCESSING-ALT-ENTRY message from the AM.

(defun atm-remove-alternative (alt-id)
  (check-type alt-id symbol)
  (setq *alternatives-table*
	(remove alt-id *alternatives-table* :key #'alt-id))
  nil)

;;; atm-get-alts-agenda-table -- called from message handling loop
;;; when received a :REQUEST-ALTS-AGENDA-TABLE message from the AM.

(defun atm-get-alts-agenda-table ()
  *alternatives-table*)

(defun atm-get-alternative (alt-id)
  (find alt-id *alternatives-table* :key #'alt-id))

(defun atm-get-number-of-alternatives ()
  (length *alternatives-table*))

(defun atm-clear-alts-agenda-table ()
  (setq *alternatives-table* nil))

;;; Alt transactions: adding alts with the ability to comit or abort
;;; Used, at present, only by KS-USER.

(defvar *alt-transaction-open-p* nil)
(defvar *saved-alternatives-table* nil)

(defun atm-begin-alt-transaction ()
  (assert (not *alt-transaction-open-p*))
  (setq *saved-alternatives-table* (copy-list *alternatives-table*)
	*alt-transaction-open-p* t)
  t)

(defun atm-commit-alt-transaction ()
  (assert *alt-transaction-open-p*)
  (setq *saved-alternatives-table* nil
	*alt-transaction-open-p* nil)
  t)

(defun atm-abort-alt-transaction ()
  (assert *alt-transaction-open-p*)
  (setq *alternatives-table* *saved-alternatives-table*
	*saved-alternatives-table* nil
	*alt-transaction-open-p* nil)
  t)

;;; ---------------------------- Change History ----------------------------
;;; (Who)   (When) 		     (What)
;;;
