;;;; File: dm-toplevel.lsp
;;; SCCS Version: %W%
;;; Contains: Top-level processing for the DM.
;;; Author: Richard Kirby (rbk) and Jeff Dalton
;;; Created: Mon Apr 16 11:29:20 1990
;;; Updated: Wed Jun 28 16:20:53 1995 by Jeff Dalton
;;; Release Version: %Y%
;;; Copyright: (c) 1992, AIAI, University of Edinburgh
;;; This material may be reproduced by or for the U.S. Government pursuant
;;; to the copyright license under the clause at DFARS 252.227-7032
;;; (June 1975) -- Rights in Technical Data and Computer Software (Foreign).

(in-package :oplan)

;; This global is used to prevent multiple :INVALIDATE-AGENDA-TABLE messages
;; being sent to the AM.
;;
(defvar *sent-invalidate-agenda-table-p* nil)

(defun dm-startup ()
  
  ;; This is the initial function for the DM process.  After startup,
  ;; it is not called again.  Instead, dm-event-handler is called
  ;; whenever a message arrives.
  
  (unless (is-init *whoami*)
    (dev-error *whoami* "is-init-failed."))
  (setf (component-init-fn *pprocess*) #'dm-init)
  (setf (component-interrupt-fn *pprocess*) #'dm-interrupt-handler)
  ;; Make breaks, etc. go to the debug window.
  (x-redirect-errors :dmout)
  ;; Register with IM.
  (dev-debug-msg "Connecting to IM via MCC." *whoami* :information)
  (clear-screen *debug-io*)
  
  (mcc-register-with-im)

  (ipc-set-event-handler 'dm-event-handler))

(defun dm-event-handler (self)
  (let ((e (next-event self)))
    (ecase (message-route e)
      (:mcc (mcc-handler (message-contents e)))
      (:triggersin
       (error ":triggersin to DM: ~S" e))
      (:dbin
       ;; DB-answer sbould be called directly.
       (error "Unexpected DM request: ~S." e)))))

(defun db-answer (ident args)
  (MONITOR :DBIN (:DETAILED "~W ~W~%" ident args))
  (let ((answer (apply (ipc-get-handler :DM ident) args)))
    (MONITOR :DBIN (:DETAILED "Answer ~W~%" answer))
    (unless *lazy-triggering*
      (check-for-triggered-entries))
    answer))

;;;; Messages

(defmessage (:DM :FUNCALL) (fn &rest args)
  (apply fn args))

(defmessage (:DM :APPLY) (fn &rest args)
  ;; last elt of args is a list of args as in apply.
  (apply #'apply fn args))

(defmessage (:DM :OBMATCH) (t1 t2 bindings)
  (ipc-with-own-copy (binds bindings (copy-tree binds))
    (obmatch3 t1 t2 binds)))

(defmessage (:DM :ADD-AGENDA-ENTRY) (ae)
  (atm-add-untriggered-agenda-entry ae)
  :OK)

(defmessage (:DM :REQUEST-CURRENT-AGENDA-TABLE) ()
  (setq *sent-invalidate-agenda-table-p* nil)
  (when *lazy-triggering*
    (check-for-triggered-entries))
  (atm-get-triggered-agenda-table))

(defmessage (:DM :PROCESSED-AGENDA-ENTRY) (ae)
  (atm-remove-triggered-agenda-entry ae)
  :OK)

(defmessage (:DM :BEGIN-ALT-TRANSACTION) ()
  (whats-going-on "Begin alt transaction")
  (atm-begin-alt-transaction))

(defmessage (:DM :COMMIT-ALT-TRANSACTION) ()
  (whats-going-on "Commit alt transaction")
  (atm-commit-alt-transaction))

(defmessage (:DM :ABORT-ALT-TRANSACTION) ()
  (whats-going-on "Abort alt transaction")
  (atm-abort-alt-transaction))

(defmessage (:DM :ADD-ALT-ENTRY) (alt)
  (prog1 (atm-add-alternative alt)
    (whats-going-on "Just added an alternative.")
    (whats-going-on "DATABASE STATE~%==============")
    (print-database)
    ))

(defmessage (:DM :REQUEST-ALTS-AGENDA-TABLE) ()
  (atm-get-alts-agenda-table))

(defmessage (:DM :PROCESSING-ALT-AGENDA) (alt-id)
  (atm-remove-alternative alt-id)
  :ok)

(defmessage (:DM :GET-EXPAND-SCHEMAS) (pat)
  (find-schemas-matching-expansion-pattern pat))

(defmessage (:DM :NUMBER-OF-EXPAND-SCHEMAS) (pat)
  (length (find-schemas-matching-expansion-pattern pat)))

(defmessage (:DM :GET-EFFECT-SCHEMAS) (effect)
  (find-schemas-for-using-for-effect effect))

(defmessage (:DM :NUMBER-OF-EFFECT-SCHEMAS) (effect)
  (length (find-schemas-for-using-for-effect effect)))

(defmessage (:DM :GET-NAMED-SCHEMA) (name)
  (find-schema-with-name name))

(defmessage (:DM :EXPAND) (node subnodes orderings)
  (ads-expand-node node subnodes orderings))

(defmessage (:DM :CHECK-TOME) (args)
  (list (apply #'tgm-add-effect args)))

(defmessage (:DM :CHECK-GOST) (cond bindings)
  (declare (ignore bindings))
  (let (act::*bindings*)
    (apply #'tgm-add-condition cond)))

(defmessage (:DM :ADD-TOME) (args)
  (list (apply #'tgm-commit-to-adding-effect args)))

(defmessage (:DM :ADD-GOST) (args)
  (list (apply #'tgm-commit-to-adding-condition args)))

(defmessage (:DM :ABORT-ADD-TOME-OR-GOST) ()
  (list (tgm-abort-adding-cond-or-effect)))

(defmessage (:DM :UPDATE-CONTRIBUTORS) (cond contributors)
  (tgm-update-contributors cond contributors))

(defmessage (:DM :ADD-LINK) (from to)
  (add-a-link from to))

(defmessage (:DM :ADD-NODE) (type pattern pre-list post-list)
  (add-a-node type pattern pre-list post-list))

(defmessage (:DM :ADD-INITIAL-NODE) (schema-node)
  (add-an-initial-node schema-node))

(defmessage (:DM :GET-PLAN-NODES) (&rest args)
  (apply #'print-plan-nodes args))

(defmessage (:DM :GET-TOME) (&optional address)
  (print-tome address))

(defmessage (:DM :GET-GOST) (&optional address)
  (print-gost address))

(defmessage (:DM :GET-RUT) ()		;resource usage table
  (rut-description))

(defmessage (:DM :DOMAIN) (filename)
  (init-new-domain filename))

;;; :SET_TASK isn't used.  See KS_SET_TASK.
; (:SET_TASK
;  (setq answer (list (set-task (car (mcc-msg-args event))))))

(defmessage (:DM :PUSH-CONTEXT) ()
  (pushcontext)
  :OK)

(defmessage (:DM :POP-CONTEXT) ()
  (popcontext)
  :OK)

(defmessage (:DM :GET-CONTEXT) ()
  (contextnum *cucontext*))

(defmessage (:DM :SET-CONTEXT) (context-number)
  (setq *cucontext* (contextwithnum context-number))
  (whats-going-on "NEW DATABASE STATE~%==================")
  (print-database)
  :OK)

(defvar *context-was-poisoned* nil)	;context-layered

(defmessage (:DM :POISON-CONTEXT) (reason)
  (setf (ctxt-symbol-value '*context-was-poisoned*) (or reason t)))

(defmessage (:DM :CONTEXT-WAS-POISONED) ()
  (ctxt-symbol-value '*context-was-poisoned*))

(defmessage (:DM :GET-VAR-VALUE) (psv)
  (psv-get-value psv))

(defmessage (:DM :GET-VAR-BINDINGS) (psv) ;call it :GET-VAR-POSSIBLES? /\/
  (psv-get-possibles psv))

(defmessage (:DM :SET-VAR-BINDING) (psv value)
  (psv-begin-var-transaction)
  (let ((answer (psv-set-value psv value)))
    (if answer				;was (eq answer t)
	(psv-commit-var-transaction)
      (psv-abort-var-transaction))
    answer))

(defmessage (:DM :ADD-VAR-RESTRICTION) (var type value)
  (psv-begin-var-transaction)
  (let ((answer (psv-add-restriction var type value)))
    (if answer				;was (eql answer t)
	(psv-commit-var-transaction)
      (psv-abort-var-transaction))
    (list answer)))

(defmessage (:DM :SET-VAR-POSSIBLES) (psv values)
  (psv-begin-var-transaction)
  (let ((answer (psv-set-possibles psv values)))
    (if answer
	(psv-commit-var-transaction)
      (psv-abort-var-transaction))
    answer))

(defmessage (:DM :EVALUATE-COMPUTE) (p v)
  (evaluate-compute-condition p v))

(defmessage (:DM :FULLY-INSTANTIABLE-P) (pat bindings)
  (fully-instantiable-p pat bindings))

(defmessage (:DM :FULLY-INSTANTIATE) (pat bindings)
  (fully-instantiate pat bindings))

(defmessage (:DM :INSTANTIATE-SCHEMA) (args)
  (db-instantiate-schema (car args) (cdr args)))

(defmessage (:DM :CHECK-ONLY-USE-IFS) (cond bindings)
  (tgm-check-condition-within-bindings cond bindings))

(defmessage (:DM :PRINT-DATABASE) ()
  (whats-going-on "DATABASE STATE~%==============")
  (print-database)
  nil)

(defmessage (:DM :CHECK-FOR-EFFECT) (effect bindings &rest optionals)
  (apply #'check-for-effect effect bindings optionals))

(defmessage (:DM :ADD-TIME-WINDOW) (from to min max)
  (add-time-window from to min max))

(defmessage (:DM :ADD-RESOURCES) (events node parent-p)
  (process-resource-events events node parent-p))

(defmessage (:DM :QA-ALL) (node-id)
  (do-qa-all node-id))

(defmessage (:DM :EXCHANGE-AGENDA-ENTRY) (old-ae new-ae)
  ;; This combines :PROCESSED-AGENDA-ENTRY and :ADD-AGENDA-ENTRY
  ;; for when we pick an alternative.
  ;; /\/: Placed here to avoid changing the time to lookup other
  ;; entries in the original case statement.  Eventually we should
  ;; put all cases in a more reasonable order...
  (atm-remove-triggered-agenda-entry old-ae)
  (atm-add-untriggered-agenda-entry new-ae)
  :OK)

(defmessage (:DM :CHECK-PLAN) (&rest args)
  (apply #'check-plan args))

(defmessage (:DM :GET-PSVS) ()
  (psv-get-psv-descriptions))

(defmessage (:DM :GET-PSVS-FOR-USER) ()
  (psv-get-psv-descriptions-for-user))
		 
(defmessage (:DM :REQUEST-UNTRIGGERED-AGENDA-TABLE) ()
  (atm-get-untriggered-agenda-table))

(defmessage (:DM :CHECK-TRIGGERS) ()	;Glen's Exec only
  ;; They're checked automatically after each request, so
  ;; this can be a no-op.
  (when *lazy-triggering*
    (check-for-triggered-entries))
  nil)

;;; End of messages


;; Set the DM to the NULL state, ie wipe everything!
(defun dm-init ()
  (clear-screen *debug-io*)
  ;; Set context to the top, and then push to a new one.
  (setq *cucontext* *globalcontext*)
  (pushcontext)
  (clear-obase)
  (ds-init)
  (clear-schema-tables)
  (clear-pop-gensymtable)
  (clear-domain-information)
  (setq tf-compiler::*user-defaults* nil) ;/\/: violates modularity
  (clear-resources)
  (qa:qa-initialise)
  (tgm:tgm-init)
  (atm-clear-agenda-tables)
  (atm-clear-alts-agenda-table)
  (psv-clear-psv-tables)
  (dev-debug-msg "initialised" *whoami* :information))

;;; ---------------------------- Change History ----------------------------
;;; (Who)   (When) 		     (What)
;;;
