;;;; File: qa-all.lsp
;;; SCCS Version: %W%
;;; Contains: qa-all and associated routines
;;; Author: Jeff Dalton
;;; Created: November 1992
;;; Updated: Fri Mar 25 03:19:03 1994 by Jeff Dalton
;;; Release Version: %Y%
;;; Copyright: (c) 1992, AIAI, University of Edinburgh
;;; This material may be reproduced by or for the U.S. Government pursuant
;;; to the copyright license under the clause at DFARS 252.227-7032
;;; (June 1975) -- Rights in Technical Data and Computer Software (Foreign).

(in-package :QA)

(export '(qa-all qa-all-at-node *qa-all-report-always-facts*))

(defparameter *qa-all-report-always-facts* t)

;;;; QA-ALL
;;;
;;; There are two entry points.  QA-ALL-AT-NODE takes a single
;;; argument, a node-end.  QA-ALL takes three arguments: pattern,
;;; value, and node-end.  In both cases, the result is a list of
;;; (pattern value) sublists.  (QA-ALL-AT-NODE n) is equivalent to
;;; (QA-ALL ?? ?? n).
;;; 
;;; The basic strategy is:
;;;  1. Get all TOME entries that are valid in the current context to
;;;     see what p=v might be true.  (N.B. this does not mean v=true.)
;;;  2. Ask QA about each possibility to see if it _is_ true.
;;;
;;; One shortcut is taken: always facts are accepted immediately,
;;; without asking QA.
;;;
;;; Note that TOME entries look like this:
;;;
;;;   (TOME <patt> <tome-at-node-end>) = <val>
;;;
;;;   <tome-at-node-end> = :ALWAYS
;;;                      | (<node> <end>)
;;;                      | (<node> <end> :ALWAYS . <bindings>)
;;;

(defun qa-all-at-node (n)
  (qa-all '??
	  '??
	  n))

(defun qa-all (p v n)
  (let ((tome-shortlist (cdr (getashortlist `(TOME ,p ??))))
	(always '())			;list of (p v) pairs
	(candidates '()))		;list of (p v) pairs
    (dolist (item tome-shortlist)
      (unless (eq (ob:value item) :undef) ;exclude unless in current context
	(let* ((a-item (psv:psv-actorise-pattern item))
	       (a-p (tgm-tome-pattern a-item))
	       (a-v (psv:psv-actorise-pattern (ob:value item))))
	  (cond ((not (match-p a-p p)))	;skip
		((not (match-p a-v v)))	;skip
		((eq (tgm-tome-node-end a-item) :always)
		 (push (list a-p a-v) always)) ;assume unique, so not pushnew
		(t
		 (pushnew (list a-p a-v) candidates :test #'equal))))))
    (let ((successes
	   (remove-if-not
	      #'(lambda (p-v)
		  (qa-w/o-work (first p-v)
			       (second p-v)
			       n))
	      candidates)))
      (if *qa-all-report-always-facts*
	  ;; Due to vars, some always facts may also be asserted elsewhere.
	  (delete-duplicates (nconc successes always) :test #'equal)
	;; Keep any re-asserted always facts.
	successes))))

(defun match-p (t1 t2)
  (let ((act::*bindings* nil))
    (ob:obmatch t1 t2)))

(defun old-qa-all (p v n)
  (let* ((tome-items (gen->list (tome-generator p v)))
	 (p-v-pairs
	  (mapcar #'(lambda (tome-item)
		      (list (tgm-tome-pattern
			     (psv:psv-actorise-pattern tome-item))
			    (psv:psv-actorise-pattern
			     (ob:value tome-item))))
		  tome-items))
	 (unique-p-v-pairs
	  (remove-duplicates p-v-pairs :test #'equal)))
    (remove-if-not
       #'(lambda (p-v)
	   (qa-w/o-work (first p-v)
			(second p-v)
			n))
       unique-p-v-pairs)))

(defun qa-w/o-work (p v n) ; -> true or false
  (let ((answer
	 (qa:qa p v n (qa:qa-make-work-flag-structure :by-bindings nil nil))))
    (not (or (null answer)
	     (some-work-to-do-p answer)))))

(defun some-work-to-do-p (or-trees)
  ;; From "kp_supportlib.lsp" /\/.
  "Returns t if there are some bindings, restrictions, or linkings to be made,
   or some deletors to remove."
  (some #'(lambda (or-tree)
	    (or (> (length or-tree) 1)
		(notevery #'(lambda (x) (eql (car x) :UPDATE))
			  (qa::or-info-rec-actions (car or-tree)))))
	or-trees))


;;;; Utilities

;;; /\/: TOME-GENERATOR should be in tgm.lsp.  In that file, VALUE,
;;; GENERATOR, and TRY-NEXT can be used w/o the package qualification.

(defun tome-generator (&optional (p '??) (v '??))
  (let ((gen (extra:generator `(TOME ,p ??) v)))
    #'(lambda ()
	(let ((item (extra:try-next gen)))
	  (loop
	    (cond ((null item)
		   ;; Gen has finished.
		   (return nil))
		  ((eq (ob:value item) :undef)
		   ;; Filter entries that have no value in the current
		   ;; context.
		   (setq item (extra:try-next gen)))
		  (t
		   (return item))))))))

;;; /\/: gen->list should be in obextra.lsp.

(defun gen->list (g)
  (do ((item (extra:try-next g)
	     (extra:try-next g))
       (result '()
	       (cons item result)))
      ((null item)
       (nreverse result))))

;;; End
