;;;; File: sanity-checker.lsp
;;; SCCS Version: %W%
;;; Contains: Consistency and reasonableness checks.
;;; Author: Richard Kirby (rbk) and Jeff Dalton
;;; Created: Wed Jul 22 12:03:49 1992
;;; Updated: Thu Jun  1 00:35:40 1995 by Jeff Dalton
;;; Release Version: %Y%
;;; Copyright: (c) 1992, 1993, 1994, AIAI, University of Edinburgh
;;; This material may be reproduced by or for the U.S. Government pursuant
;;; to the copyright license under the clause at DFARS 252.227-7032
;;; (June 1975) -- Rights in Technical Data and Computer Software (Foreign).

(in-package :oplan)

;;; The sanity-checker is Used after the planner has claimed that it
;;; has finished to check that the plan is indeed consistent and that
;;; various data structures are as they should be.
;;;
;;; To request a sanity check, type the following to the REPL:
;;;
;;;   (send-to-leftin-if-valid '(:check-plan))
;;;
;;; or the equivalent
;;;
;;;   (sanity-check)
;;;
;;; A description of the tests performed and the results obtained will
;;; be printed in the Lisp interaction window.  When all tests are complete,
;;; the total number of errors will be reported in the TA window.
;;;
;;; The sanity-checker can also be run by the auto-tester by specifying
;;; either
;;;
;;;   :sanity-check t
;;;
;;; or
;;;
;;;   :sanity-check-only t
;;;
;;; when running a test sequence.
;;;

;;; /\/: Counting the errors should be part of check-error.

;;; /\/: Some more checks to do:
;;;
;;;  * Node ends have unique indexes.
;;;  * The :begin of each node is linked before the :end.
;;;  * Nodes are linked between the ends of their parents.
;;;  * Contributors are before the condition node-ends
;;;  * PSV not-sames are obeyed.
;;;  * Begin_of node-1 has no predecessors, end_of node-2 no successors,
;;;    and no other node-ends are without both predecessors and successors.
;;;  * There are no cycles in the node orderings.  This can be checked
;;;    by calling node-ends-between on begin_of node-1 and end_of node-2.
;;;    Then check that all node-ends are in the node-ends-between result.
;;;  * Recomputing the TPN from scratch produces the same values.
;;;  * Consumable resources are within the original "overall" bounds.
;;;  * Simulated execution that tracks condition and effect pv-pairs.
;;;


;;; Check definitions, etc.

(defvar *all-checks* '())

(defvar *check* 'sanity-checker)	;check now running

(defvar *check-number* 0)		;increases by one for each check

(defvar *check-out* (make-synonym-stream '*terminal-io*))

(defmacro define-check (name parameters description &rest body)
  (check-type name symbol)
  (check-type description string)
  `(progn
     (add-check ',name)
     (setf (check-description ',name) ',description)
     (defun ,name ,parameters
       ,@body)))

(defmacro check-description (check) `(get ,check 'check-description))

(defun add-check (name)
  (unless (member name *all-checks*)
    (setq *all-checks* (nconc *all-checks* (list name)))))


;;; The sanity-checker.

(defun check-plan ()			;called from db-answer
  (sanity-checker))

(defun sanity-checker (&optional (*check-out* *check-out*))
  (unless (eq (pprocess-name *pprocess*) :dm)
    (error "The sanity checker must run in the DM, not in the ~S process."
	   (pprocess-name *pprocess*)))
  (clear-check-identification)
  (let ((*print-case* :downcase)
	(*check-number* 1)
	(total-errors 0))
    (dolist (*check* *all-checks*)
      (let ((errors (or (funcall *check*) 0)))
	(when (> errors 0)
	  (check-report "Check ~D -- ~D errors" *check-number* errors))
	(incf total-errors errors)
	(incf *check-number*)))
    (when (> total-errors 0)
      (check-report "Total errors = ~D" total-errors))
    total-errors))
  

;;; Check reports

(defun check-report (format-string &rest format-args)
  (ensure-check-is-identified)
  (mon-format *check-out* "~&~?~%" format-string format-args))

(defun check-error (format-string &rest format-args)
  (check-report "Error: ~?" format-string format-args))

(let ((checker-identified-p nil)
      (last-identified-check nil))

  (defun clear-check-identification ()
    (setq checker-identified-p nil
          last-identified-check nil))

  (defun ensure-check-is-identified ()
    (unless checker-identified-p
      (format *check-out* "~&~%SANITY CHECKER:~%")
      (setq checker-identified-p t))
    (unless (or (eq *check* last-identified-check)
		(eq *check* 'sanity-checker))
      (format *check-out* "~&~%Check ~D ~A:~%[~A]:~%"
	      *check-number* *check* (check-description *check*))
      (setq last-identified-check
	    *check*))))


;;;; Sanity checks

(define-check check-for-poisoned-alts ()
  "Alternatives in poisoned contexts"
  (let ((errors 0))
    (dolist (alt *alternatives-table*)
      (incontext (contextwithnum (alt-context alt))
	#'(lambda ()
	    (let ((reason (ctxt-symbol-value '*context-was-poisoned*)))
	      (when reason
		(incf errors)
		(check-error "Alt ~S in poisoned context ~S~%~
                              Reason: ~S"
			     (alt-id alt) (alt-context alt) reason))))))
    errors))

;;; Foreach TOME entry
;;;   Check that the effect node-end is in the current context.

(define-check check-effect-node-end-context ()
  "TOME node-ends in present contex"
  (let ((errors 0))
    (map-over-tome
      #'(lambda (entry value)
	  (let ((node-end (tgm-tome-node-end entry)))
	    (unless (node-end-in-present-context-p node-end)
	      (incf errors)
	      (check-error "~A = ~A:~%~A not in this context"
			   entry value
			   (ds-node-tag node-end))))))
    errors))

;;; Foreach GOST entry
;;;   Check the condition node-end is in the current context.

(define-check check-condition-node-end-context ()
  "GOST node-ends in present context"
  (let ((errors 0))
    (map-over-gost
      #'(lambda (entry value)
	  (let ((node-end (tgm-gost-node-end entry)))
	    (unless (node-end-in-present-context-p node-end)
	      (incf errors)
	      (check-error "~A = ~A:~%~A not in this context."
			   entry value
			   (ds-node-tag node-end))))))
    errors))

;;; Foreach GOST entry
;;;   Get the list of contributors.
;;;   Check each contributor node-end is in the current context.
;;;   If nil or :NONE then a problem

(define-check check-contributors-node-end-contexts ()
  "GOST contributors node-ends in present context"
  (let ((errors 0))
    (map-over-gost
      #'(lambda (entry value)
	  (let ((bad-node-ends
		 (remove-if
		   #'(lambda (contributor)
		       (node-end-in-present-context-p (car contributor)))
		   value)))
	    (when bad-node-ends
	      (incf errors (length bad-node-ends))
	      (check-error "~A = ~A:~%~A not present in this context"
			   entry value bad-node-ends)))))
    errors))

;;; Foreach GOST entry
;;;   Foreach contributor
;;;     Check that a TOME entry with pattern matching the condition is present.

(define-check check-that-contributors-provide-effects ()
  "GOST contributors node-ends provide effect"
  (let ((errors 0))
    (map-over-gost
      #'(lambda (entry value)
	  (let ((cond-type (tgm-gost-condition-type entry))
		(cond-pattern (tgm-gost-pattern entry))
		(cond-value (tgm-gost-value entry))
		(missing-contributors '())
		(bad-contributors '()))
	    (dolist (contributor value)
	      (let* ((node-end (car contributor))
		     (tome-entry (tgm-make-tome-entry cond-pattern node-end))
		     (tome-value (value tome-entry)))
		(if (eq cond-type 'supervised)
		    (unless (node-or-child-provides-effect-p
			     node-end cond-pattern cond-value)
		      (incf errors)
		      (push tome-entry missing-contributors))
		  ;; Not supervised
		  (ecase (node-provides-effect-p
			  node-end cond-pattern cond-value)
		    ((nil)
		     (incf errors)
		     (push tome-entry missing-contributors))
		    ((:ok))
		    ((:deletes)
		     (incf errors)
		     (push (list node-end tome-value) bad-contributors))))))
	    (when missing-contributors
	      (check-error
	          "~A = ~A:~%~
                   ~{~6T ~A not present in this context~%~}"
		  entry value (reverse missing-contributors)))
	    (when bad-contributors
	      (check-error
		  "~A = ~A:~%~
                   ~6T Contributors with the wrong effect value:~
                   ~6T   ~:{~A --> ~A~%~}"
		  entry value (reverse bad-contributors))))))
    errors))

;; Node-provides-effect-p returns:
;;   nil if the node-end has no effect that matches cond-pattern,
;;   :ok if the node-end has an effect that matches both cond-pattern
;;          and cond-value,
;;   :deletes if the node-end has an effect that matches cond-pattern
;;               but not cond-value.

(defun node-provides-effect-p (node-end cond-pattern cond-value)
  (map-over-matching-tome
    cond-pattern
    #'(lambda (entry value)
	(when (equal (tgm-tome-node-end entry) node-end)
	  (return-from node-provides-effect-p
	    (if (obmatch3 value cond-value nil)
		:ok
		:deletes)))))
  nil)

(defun node-or-child-provides-effect-p (node-end cond-pattern cond-value)
  ;; Called only for supervised conditions.
  ;; Make sure the node-end isn't something unexpected.
  (assert (and (consp node-end) (null (cddr node-end))))
  (let ((node-name (symbol-name (etag-node node-end))))
    (map-over-matching-tome
       cond-pattern
       #'(lambda (entry value)
	   (let ((tome-node-name
		  (symbol-name (etag-node (tgm-tome-node-end entry)))))
	     (when (and (eql 0 (search node-name tome-node-name))
			(obmatch3 value cond-value nil))
	       (return-from node-or-child-provides-effect-p
		 t)))))
    nil))

;;; Foreach GOST entry
;;;   Foreach TOME entry which matches the condition pattern
;;;                            but differs in value
;;;     Check that the TOME entry is not within the ranges set up
;;;                by the contributors.

(define-check check-that-deleters-cannot-delete ()
  "Deleters out of GOST ranges"
  (let ((errors 0))
    (map-over-gost
      #'(lambda (entry value)
	  (let ((cond-pattern (tgm-gost-pattern entry))
		(cond-value (tgm-gost-value entry))
		(cond-node-end (tgm-gost-node-end entry))
		(cond-contributors value)
		(bad-deleters '()))
	    (map-over-tome
	      #'(lambda (entry value)
		  (let* ((tome-pattern (tgm-tome-pattern entry))
			 (tome-node-end (tgm-tome-node-end entry))
			 (tome-value value)
			 (bindings (obmatch3 tome-pattern cond-pattern nil)))
		    ;; If this TOME entry conflicts with the GOST entry
		    ;; we are considering, then check that it is outside
		    ;; the range.
		    (when (and bindings
			       (not (obmatch3 tome-value cond-value bindings)))
		      (unless (or (equal tome-node-end cond-node-end)
				  (not (tgm::tgm-node-end-within-range-p
					tome-node-end cond-node-end
					cond-contributors)))
			(incf errors)
			(push (cons entry value) bad-deleters))))))
	    (dolist (deleter bad-deleters)
	      (check-error
	         "~A = ~A~%is within the range of~%~A = ~A"
		 (car deleter) (cdr deleter) entry value)))))
    errors))

;;; Foreach PSV
;;;   Check that the PSV has a value.

(define-check check-psv-values ()
  "Check that all PSVs have a value"
  (let ((errors 0))
    (flet ((check-psv-has-value (tag body)
	     (when (eql (psv::psv-body-value body) :undef)
	       (incf errors)
	       (check-error "~A has no value." tag))))
      (map-over-psvs #'check-psv-has-value))
    errors))

;;; Foreach PSV
;;;   Check that not-sames are obeyed.

(define-check check-psv-not-sames ()
  "Check that all PSV not-sames lists are obeyed"
  (let ((errors 0))
    (map-over-psvs
      #'(lambda (tag body)
	  (let ((val (psv::psv-body-value body)))
	    (mapc #'(lambda (ns-tag)
		      (when (eql val (psv-get-value ns-tag))
			(incf errors)
			(check-error "~A and ~A = ~A despite a not-same"
				     tag ns-tag val)))
		  (psv::psv-body-not-sames body)))))
    errors))


;;; Utilities

;; (map-over-tome fn) --
;; Applies fn to each TOME entry that is in the current context.
;; fn takes two arguments: the TOME entry, and the TOME value.

(defun map-over-tome (fn)
  (let ((tome-entries (generator '(TOME ?? ??) '??))
	tome-value)
    (do ((tome-entry (try-next tome-entries) (try-next tome-entries)))
	((null tome-entry))
      (setq tome-value (value tome-entry))
      (setq tome-entry (psv-actorise-pattern tome-entry))
      (unless (eq tome-value :undef)
	(funcall fn tome-entry tome-value)))))

(defun map-over-matching-tome (pat fn)
  (assert (equal pat (psv-actorise-pattern pat)))
  (let ((tome-entries (generator '(TOME ?? ??) '??))
	tome-value)
    (do ((tome-entry (try-next tome-entries) (try-next tome-entries)))
	((null tome-entry))
      (setq tome-value (value tome-entry))
      (setq tome-entry (psv-actorise-pattern tome-entry))
      (unless (or (eq tome-value :undef)
		  (not (obmatch3 pat (tgm-tome-pattern tome-entry) nil)))
	;; /\/: Do we need to pass along the bindings from the obmatch?
	(funcall fn tome-entry tome-value)))))

;; (map-over-gost fn) --
;; Applies fn to each GOST entry that is in the current context.
;; fn takes two arguments: the GOST entry, and the GOST value.

(defun map-over-gost (fn)
  (let ((gost-entries (generator '(GOST ?? ?? ?? ??) '??))
	gost-value)
    (do ((gost-entry (try-next gost-entries) (try-next gost-entries)))
	((null gost-entry))
      (setq gost-value (value gost-entry))
      (setq gost-entry (psv-actorise-pattern gost-entry))
      (unless (eql gost-value :undef)
	(funcall fn gost-entry gost-value)))))

(defun map-over-psvs (fn)
  (maphash #'(lambda (tag body-ctxt-list)
	       (let ((body (derefincontext body-ctxt-list)))
		 (when body
		   (funcall fn tag body))))
	   psv::*psv-table*))

(defun node-end-in-present-context-p (node-end)
  (or (eql node-end :ALWAYS)
      (let ((node-tag (ds-node-tag node-end)))
	(typep (get-node node-tag)
	       'ads-node))))

;;; ---------------------------- Change History ----------------------------
;;; (Who)   (When) 		     (What)
;;;
