;;;; File: complr.lsp
;;; SCCS Version: %W%
;;; Contains: Top level of the TF compiler
;;; Author: Jeff Dalton <J.Dalton@ed.ac.uk>
;;; Created: March 1992
;;; Updated: Tue Jul  4 00:11:59 1995 by Jeff Dalton
;;; Release Version: %Y%
;;; Copyright: (c) 1992, 1993, 1994, 1995 AIAI, University of Edinburgh
;;; This material may be reproduced by or for the U.S. Government pursuant
;;; to the copyright license under the clause at DFARS 252.227-7032
;;; (June 1975) -- Rights in Technical Data and Computer Software (Foreign).

(in-package :tf-compiler)

;;; Note that there are many places where it looks like we could use
;;; SET-DIFFERENCE.  But we can't (or don't want to) because SET-DIFFERENCE
;;; doesn't guarantee the order of the result.  Consequently, we've defined
;;; a function STABLE-SET-DIFFERENCE.

;;; /\/: It may be better to interleave the two main passes rather
;;; than have such things as the undefined type test happen well after
;;; parsing.  (See how easy the node ref check is, eg. -- in the parser.)
;;; The "semantic" errors might then have line number refs too.


;;; Entry points

;;; Compile-tf-stream is the main entry point to the compiler.

;;; /\/: But the DM actually calls compile-from-tf-file.

;;; /\/: Some comments are needed.

;;; /\/: The check-only parameter isn't necessary a good idea.
;;; It causes the compiler to stop after parsing, but a number of
;;; errors aren't detected until later.  So if you really want to
;;; check a domain definition, you should not use :check-only t.

;;; /\/: It may be best to return NIL as the first value when there
;;; have been errors.

(defun compile-tf-stream (stream &optional (input-domain (make-domain))
				           (check-only nil))
  (handler-bind
      ((error
	#'(lambda (c)
	    (dev-debug :force-out "~&Lisp error during compilation:~%~A~%" c)
	    (case (menu-request
		   `("-heading" "Lisp error during compilation"
		     "Ok=:ok"
		     "Break 1st=:break"))
	      (:ok)
	      (:break (break "TF compiler error")))
	    (return-from compile-tf-stream
	      (values input-domain
		      1			;we only _know_ about 1 /\/
		      nil)))))
    (do-compile-tf-stream stream input-domain check-only)))

(defun do-compile-tf-stream (stream input-domain check-only)
  ;; Parse
  (let* ((output-domain (copy-domain input-domain))
	 (*input-domain* input-domain)
	 (*output-domain* output-domain)
	 (*domain* output-domain)
	 (*tf-stream* stream)
	 (*syntax-check-only* check-only)
	 (*error-count* 0))

    (set-domain-name output-domain stream)
    (multiple-value-bind (parse-tree syntax-errors)
	(compile-tf #'<tf-file> (tf-tokens *tf-stream*))
      (when check-only
	(return-from do-compile-tf-stream
	  (values input-domain
		  syntax-errors
		  nil)))

      ;; Generate domain information from the parse tree.
      (construct-domain output-domain parse-tree)

      ;; Produce various tables, etc.
      (analyze-domain output-domain)

      ;; If there are errors, return the domain we were given
      ;; without any modifications.  Otherwise return the modified
      ;; domain.
      (let* ((semantic-errors *error-count*)
	     (total-errors (+ syntax-errors semantic-errors)))
	(if (> total-errors 0)
	    (values input-domain total-errors '())
	    (values output-domain
		    total-errors)))))) ;s.b. zero

(defun set-domain-name (domain stream)
  (let ((name (handler-case (pathname-name stream) (error () "anonymous"))))
    (setf (domain-name domain)
	  (if (domain-name domain)
	      (concat-string (domain-name domain) "+" name)
	    name))))

;;; For debugging (and other things), we can compile from a string
;;; or a file.

(defun compile-from-tf-string (string &rest other-args)
  (with-open-stream (s (make-string-input-stream string))
    (apply #'compile-tf-stream
	   s
	   other-args)))

(defun compile-from-tf-file (file &rest other-args)
  (with-open-file (s file :direction :input)
    (apply #'compile-tf-stream
	   s
	   other-args)))


;;; Test files (or, rather, syntax-check them)

;;; Compile-test-files can be used to compile all the ".tf" files in a
;;; directory.  It prints a report of the number of errors in each file
;;; at the end.

;;; Note that you may have to do some work to make sure it interprets
;;; its argument as a directory.  Eg, in KCL you might have to end the
;;; name with a "/", as in: (compile-test-files "tf-examples/").

(defun compile-test-files (directory &key (messages nil))
  (multiple-value-bind (pattern filenames) (directory-files directory "tf")
    (let ((*suppress-messages* (not messages))
	  (error-report '()))
      (dolist (file filenames)
	(when messages
	  (format t "~&--------------------------------------------------~%"))
	(format t "~S:~%" (namestring file))
	(multiple-value-bind (result errors)
	    (compile-from-tf-file (merge-pathnames file pattern))
	  (declare (ignore result))
	  (push (list file errors)
		error-report)))

      ;; Report the number of errors for each file
      (format t "~&--------------------------------------------------~%")
      (format t "~&File ~30T Errors")
      (format t "~&---- ~30T ------")
      (dolist (r (reverse error-report))
	(let ((file (first r))
	      (errors (second r)))
	  (if (> errors 0)
	      (format t "~&~A ~30T ~6D" (namestring file) errors)
	      (format t "~&~A"          (namestring file)))))
      (format t "~%")))
  (values))

(defun directory-files (directory type)
  (let* ((pattern
	  (merge-pathnames (make-pathname :name :wild
					  :type type)
			   (truename directory)))
	 (filenames
	  (mapcar #'(lambda (name)
		      ;; /\/: (enough-namestring name pattern)
		      ;; loses if the name is something like a.b.tf.
		      (make-pathname :name (pathname-name name)))
		  (directory pattern))))
    (values pattern filenames)))

;;; End
