;;;; File: triggers.lsp
;;; SCCS Version: %W%
;;; Contains: Code for handling agenda entry triggering.
;;; Author: Richard Kirby (rbk)
;;; Created: Thu Jul 11 13:57:25 1991
;;; Updated: Thu Aug 11 17:05:15 1994 by Jeff Dalton
;;; Release Version: %Y%
;;; Copyright: (c) 1992, AIAI, University of Edinburgh
;;; This material may be reproduced by or for the U.S. Government pursuant
;;; to the copyright license under the clause at DFARS 252.227-7032
;;; (June 1975) -- Rights in Technical Data and Computer Software (Foreign).

#| This is kept very simple at the moment. Each time the DM handles a query, it
calls check-for-triggered-entries which then looks through all the untriggered
agenda entries and checks to see if their triggers are triggered. This will be
improved to be reactive, eg when bind a variable then check the bind triggers
to see if any agenda entries are now triggered - not check any and every
trigger!
|#

(in-package :oplan)

(defparameter *lazy-triggering* t)	;/\/ to let us test this w/o commit

(defun check-for-triggered-entries ()
  "Checks for triggered entries. If there are any, then add them to the
   triggered agenda table, and then notify the AM that its copy is out of
   date."
  (declare (special *sent-invalidate-agenda-table-p*))
  (let ((table (atm-get-untriggered-agenda-table))
	triggered-entries ae)
    (declare (special ae))		;/\/ lose, lose.
    (dolist (entry table)
      (setq ae entry)
      (if (is-triggered-p (ag-trigger entry))
	  (push entry triggered-entries)))
    (dolist (entry triggered-entries)
      (atm-remove-untriggered-agenda-entry entry)
      (atm-add-triggered-agenda-entry entry))
    ;; Notify the AM that it needs to fetch a new copy of the current Agenda
    ;; Table.
    (if triggered-entries
	(unless (or *lazy-triggering* *sent-invalidate-agenda-table-p*)
	  (ipc-send :TRIGGERSIN :INVALIDATE-AGENDA-TABLE)
	  (setq *sent-invalidate-agenda-table-p* t)))))

(defun is-bound-p (psv-actor)
  (psv-has-value-p psv-actor))

(defun is-triggered-p (trigger)
  (declare (special ae))		;/\/ lose, lose.
  (cond ((eq trigger t)
	 t)
	((eq (car trigger) :BIND)
	 (is-bound-p (cdr trigger)))
	((eq (car trigger) :AND)
	 (every #'is-triggered-p (cdr trigger)))
	((eq (car trigger) :OR)
	 (some #'is-triggered-p (cdr trigger)))
	((eq (car trigger) :EMPTY)
	 ;; Check to see if the agenda-table is empty and only one thing on
	 ;; the untriggered agenda table (this entry), in which case we
	 ;; assume the planner has completed.
	 (atm-empty-tables-p))
	((eq (car trigger) :WAIT-ON-EFFECT)
	 ;; Checks that there are no EXPAND and ACHIEVE triggered
	 ;; agenda entries that could possibly cause the effect
	 ;; specified to be added.
	 (not (check-for-effect (cdr trigger) nil (ag-id ae))))
	(t
	 ;; trigger might be a fact, so lets check it.
	 ;; Look for at least one gost entry (irrespective of condition type)
	 ;; which has a value other than :undef.
	 (let ((gen (generator (list* 'gost actorsym trigger) actorsym)))
	   (do ((entry (try-next gen) (try-next gen)))
	       ((null entry) nil)
	     (unless (eq (value entry) :undef)
	       (return t)))))))

(defun check-for-effect (effect bindings &optional ae-being-checked)
  ;; Checks to see if the agenda-table has any EXPAND or ACHIEVE agenda entries
  ;; that could possibly cause the effect required to be added to the plan
  ;; state.
  ;; NOTE: Should we also check the untriggered agenda table? But if so, we
  ;; might need to exclude checking a particular agenda entry (namely if this
  ;; is called from is-triggered-p then the agenda entry with this trigger).
  (declare (ignore bindings))
  (dolist (ae (atm-get-triggered-agenda-table))
    (unless (eql (ag-id ae) ae-being-checked)
      (let ((ks-name (car (ag-body ae))))
	(if (eql ks-name :EXPAND)
	    (if (member (car effect)
			(get-possible-effects-from-expanding
			 (caddr (ag-body ae))))
		(return-from check-for-effect t))
	    (if (eql ks-name :ACHIEVE)
		(if (member (car effect)
			    (get-possible-effects-from-achieving
			     (con-pattern (cadr (ag-body ae)))))
		    (return-from check-for-effect t))
		(if (eql ks-name :OR)
		    (let ((first-action (car (qa::or-info-rec-actions
					      (car (fifth (ag-body ae)))))))
		      (if (eql (car first-action) :REACHIEVE)
			  (let ((cond (cdadr first-action)))
			    (if (and (eql (con-type cond) 'achievable)
				     (member
				      (car effect)
				      (get-possible-effects-from-achieving
				       (con-pattern cond))))
				(return-from check-for-effect t)))))))))))
  (unless (null (atm-get-triggered-agenda-table))
    (dolist (ae (atm-get-untriggered-agenda-table))
      (unless (eql (ag-id ae) ae-being-checked)
	(let ((ks-name (car (ag-body ae))))
	  (if (eql ks-name :EXPAND)
	      (if (member (car effect)
			  (get-possible-effects-from-expanding
			   (caddr (ag-body ae))))
		  (return-from check-for-effect t))
	      (if (eql ks-name :ACHIEVE)
		  (if (member (car effect)
			      (get-possible-effects-from-achieving
			       (con-pattern (cadr (ag-body ae)))))
		      (return-from check-for-effect t))
		  (if (eql ks-name :OR)
		      (let ((first-action (car (qa::or-info-rec-actions
						(car (fifth (ag-body ae)))))))
			(if (eql (car first-action) :REACHIEVE)
			    (let ((cond (cdadr first-action)))
			      (if (and (eql (con-type cond) 'achievable)
				       (member
					(car effect)
					(get-possible-effects-from-achieving
					 (con-pattern cond))))
				  (return-from check-for-effect t)))))))))))))

;;;; ---------------------------- Change History ----------------------------
;;;; (Initials) (Date of change)           (Comment)
