;;;; File: mcc-server.lsp
;;; SCCS Version: %W%
;;; Contains: Module Communication Channel code specific to the IM process.
;;; Author: Richard Kirby (rbk) and Jeff Dalton
;;; Created: Thu Sep 13 11:51:54 1990
;;; Updated: Thu Jun 22 19:33:28 1995 by Jeff Dalton
;;; Release Version: %Y%
;;; Copyright: (c) 1992, 1993 AIAI, University of Edinburgh
;;; This material may be reproduced by or for the U.S. Government pursuant
;;; to the copyright license under the clause at DFARS 252.227-7032
;;; (June 1975) -- Rights in Technical Data and Computer Software (Foreign).

(in-package :MCC)

;; mcc-defs defines the mcc package /\/
;(use-package '(:DEV :UTIL :IPC :COMPONENTS))
;(use-package :MONITORS)

;; But we can't resolve this circularity there /\/
(use-package :IPC)

(import '(oplan::make-agenda-rec))

(export '(mcc-controller-handler))

(defvar *mcc-connection-list* nil)

;;; Handler for the server end of :MCC.

;;; N.B. We don't use DEFMESSAGE here because some messages look at
;;; more than the mcc-msg-ident and -args of the message, which would
;;; be inconsistent with other uses of DEFMESSAGE, and because not all
;;; messages look at everything, which would require lots of IGNORE
;;; declarations.

(defun mcc-controller-handler (stream event)
  (ecase (mcc-msg-ident event)

    (:DYING
     (mcc-handle-dying-system event))

    (:OK
     ;; Should never receive an :OK at this level.
     (dev-debug :non-fatal-error "Unexpected :OK message received: ~S." event))

    (:REGISTER
     (dev-debug :information "~S just registered." (mcc-msg-sender event))
     (push (cons (mcc-msg-sender event) stream) *mcc-connection-list*)
     #+:undef
     (ipc-write stream (mcc-make-message :MCC :OK)))

    (:RESULT
     (destructuring-bind (form result) (mcc-msg-args event)
       (dev-debug :user-request "Result of ~S is ~S" form result)))

    (:PRINT
     (let ((channel (first (mcc-msg-args event)))
	   (msg (rest (mcc-msg-args event))))
       (monitor :IM
	 (:DETAILED "Printing to ~S: ~S" channel msg))
       ;; :RIGHTOUT is a special case only because we may have to start
       ;; up the agent to the right since it may not be running yet.
       (case channel
	 (:RIGHTOUT
	  (send-to-agent-on-right msg))
	 (otherwise
	  (ipc-write channel msg)))))

    (:INIT
     ;; Sends INIT messages to all connected processes, and then returns
     ;; an OK message to the sender of this message.
     (mcc-init-planner (mcc-msg-sender event))
     #+:undef
     (ipc-write stream (mcc-make-message :MCC :OK)))

    (:SET-DEBUG-LEVEL
     ;; Sends :DEBUG-LEVEL message to the process/processes requested.
     (mcc-send-debug-level event))

    (:SINGLE-STEP
     ;; Sends :SINGLE-STEP message to a process.
     (destructuring-bind (destination value) (mcc-msg-args event)
       (assert (eq :MCC (cdr (assoc destination *mcc-connection-list*))))
       (ipc-send-priority-message
	  :MCC destination (mcc-make-message :MCC :SINGLE-STEP value))))

    (:INTERRUPT
     ;; Sends :INTERRUPT message to a process.
     (let ((destination (car (mcc-msg-args event))))
       (assert (eq :MCC (cdr (assoc destination *mcc-connection-list*))))
       (if (eq destination :KP-1)
	   (ipc-send-priority-message :AGENDAIN :AM
	     (list :EVENT
		   (make-agenda-rec
		     :trigger t
		     :body '(:USER :USER-REQUEST))))
	 (ipc-send-priority-message
	   :MCC destination (mcc-make-message :MCC :INTERRUPT)))))

    (:SEND-TO-VIEWER
     (send-to-plan-viewer
      (car (mcc-msg-args event))))

    (:DIARY
     (dev-debug-msg (format nil "Sending ~S to DIARY" (mcc-msg-args event))
		    *whoami* :information)
     (ipc-write :DIARYIN event))

    ;; ADD MORE HERE.
    ))


;;; :SET-DEBUG-LEVEL

(defun mcc-send-debug-level (event)
  (destructuring-bind (who level) (mcc-msg-args event)
    (setq level (or (dev::dev-get-level-num-from-level-name level)
		    10))
    (case who
      ((:ALL)
       (dolist (proc *mcc-connection-list*)
	 (ipc-send-priority-message
	   (cdr proc) (car proc) (mcc-make-message :MCC :DEBUG-LEVEL level)))
       (setq dev::*dev-debug-level* level))
      ((:IM)
       (dev-set-debug-level level))
      (otherwise
       (ipc-send-priority-message
	  (cdr (assoc who *mcc-connection-list*))
	  who
	  (mcc-make-message :MCC :DEBUG-LEVEL level))))))


;;; :INIT

;; Sends an INIT message to each connected process, except the KP that sent
;; the INIT message, and calls an init function.
(defun mcc-init-planner (sender)
  ;; Init self.
  (funcall (component-init-fn *pprocess*))
  ;; Typical connection-list: ((:KP-1 . :MCC) (:DM . :MCC) (:AM . :MCC))
  ;; Typical sender: :KP-1
  (dolist (proc *mcc-connection-list*)
    (unless (eq (car proc) sender)
      (dev-debug-msg (format nil "Sending :INIT to ~A" (car proc))
		     *whoami* :information)
      (ipc-write-client (cdr proc) (car proc) (mcc-make-message :MCC :INIT))
      (mcc-wait-for-ack (cdr proc)))))


;;; :DYING

;;; /\/ We ought to send messages to any connected agents and any
;;; processes that aren't on the connection list (eg, random p-processes
;;; in the single-process version).  But it's better to do something
;;; that at least gets this agent to exit rather than wait until we
;;; have time to do it right.  Hence the call to exit-oplan.

(defun mcc-handle-dying-system (who)
  (dev-debug-msg (format nil "~A is dying" (mcc-msg-sender who))
		 *whoami* :emergency)
  ;; If there's a separate agent to the right, tell it to exit.
  (mcc-kill-agent-to-right)
  ;; Forget the process that started us exiting.
  (setq *mcc-connection-list* (remove (mcc-msg-sender who)
				      *mcc-connection-list* :key #'car))
  ;; Now tell all the other processes to die. Wait for
  ;; the results, and then exit myself.
  (mcc-kill-off-planner *mcc-connection-list*)
  (oplan::exit-oplan)			;/\/ Out, out, out!
  (ipc-terminate))

(defun mcc-kill-off-planner (list-of-processes)
  "Sends a :DIE-DIE-DIE message to each process that the IM currently knows
   about."
  (dolist (proc list-of-processes)
    (dev-debug-msg (format nil "Sending :DIE-DIE-DIE to ~A" (car proc))
		   *whoami* :emergency)
    (ipc-write-client (cdr proc) (car proc)
		      (mcc-make-message :MCC :DIE-DIE-DIE))
    (mcc-wait-for-ack (cdr proc))))

(defun mcc-kill-agent-to-right ()
  (dolist (possibility '(:world-io :right-io))
    (let ((p (exists-pprocess possibility)))
      (when p
	(send-to-pprocess p '(:kill))
	;; Transmit manually, since we'll exit before the pprocess runs.
	(pprocess::transmit-from-socket-to-stream p)
	(return)))))


;;; Wait-for-ack

#|
;; An IM specific version.
(defun mcc-wait-for-ack (&optional (channel :mcc))
  "Waits for a :OK message on the given channel."
  (dev-debug-msg (format nil "Got :OK~%") *whoami* :information))
|#


;;; ---------------------------- Change History ----------------------------
;;; (Who)   (When) 		     (What)
;;;
