;;;; File: KS-BIND.lsp
;;; SCCS Version: %W%
;;; Contains: Bind up variables.
;;; Author: Jeff Dalton and Richard Kirby (rbk)
;;; Created: Mon Jul 22 16:25:23 1991
;;; Updated: Thu Sep 29 18:31:16 1994 by Jeff Dalton
;;; Release Version: %Y%
;;; Copyright: (c) 1992, 1994 AIAI, University of Edinburgh
;;; This material may be reproduced by or for the U.S. Government pursuant
;;; to the copyright license under the clause at DFARS 252.227-7032
;;; (June 1975) -- Rights in Technical Data and Computer Software (Foreign).

(in-package :knowledge-source)

(defun KS-BIND (ag)
  (STAGE-MANAGER
      (bindings			;possible values in an alternative
       force-auto)		;true in an alt that shouldn't ask the user
      (var-name)
      ag

   ;; There's only one stage, but we still save some state for
   ;; use in alternatives.

   (setq var-name (cadr (ag-body ag)))

   ;; If bindings is set then this is from an alternative invocation, so
   ;; no need to get the possibles.
   (unless bindings
     ;; Ask the DM for possible bindings.
     (MONITOR :KP-1 (:DETAILED "Asking DM for bindings for ~S~%" var-name))
     (setq bindings (get-bindings var-name)))

   ;; Now bindings is either :already-bound or a list of possible vaues.
   (cond ((eql bindings :already-bound)
	  (let ((value (db-request :get-var-value var-name)))
	    (MONITOR :KP-1
	     (:DETAILED "~S is already bound to ~S~%" var-name value))
	    (whats-going-on "Variable ~S is already bound." var-name value)))
	 ((null bindings)
	  (whats-going-on "No possible bindings for ~S, so POISON PLAN."
			  var-name)
	  (post-agenda `(:POISON-STATE :COULD-NOT-FIND-BINDING ,var-name)))
	 ((not (consp bindings))
	  (error "Strange value for bindings ~S." bindings))
	 ((length=1 bindings)		; (= (length bindings) 1)
	  (set-binding-else-poison var-name (car bindings)))
	 ;; Now know (> (length bindings) 1)
	 ((or force-auto (eq *psv-binding-mode* :auto))
	  (auto-bind ag var-name bindings))
	 (t
	  ;; Ask the user via KS-USER.
	  ;; Set the possibles cache so it looks right and so
	  ;; KS-USER sees only the values that are left.
	  ;; Note that we're going to set the cache to contain
	  ;; more than one value, because otherwise we wouldn't
	  ;; need to ask the user.
	  ;; /\/: If this fails, someone had better investigate.
	  ;; /\/: It wouldn't be necessary if the user never saw
	  ;;      any of the :BINDs that came from alternatives.
	  ;;      That is, if in alts we passed on force-auto.
	  (assert (db-request :set-var-possibles var-name bindings))
          ;; Post a :USER entry.			      
	  ;; /\/: Should we use the agent agenda?  Or can :USER be on either?
	  (post-agenda
	   `(:USER :SYSTEM-REQUEST
	           :BIND ,var-name))))))

(defun auto-bind (ag var-name bindings)
  (let ((choice (car bindings))
	(other-values (cdr bindings)))

    ;; When we first enter this function, bindings contains more than
    ;; one value, but later on it may not.
    (when (null other-values)
      (set-binding-else-poison var-name choice)
      (return-from auto-bind))
    
    (MONITOR :KP-1 (:ONE-LINER "Binding ~S to ~S~%" var-name choice)
	           (:DETAILED "Automatically using first choice.~%~
                               Binding ~S to ~S~%~
                               ~S are the alternatives.~%"
			      var-name choice other-values))

    (whats-going-on "More than one applicable binding for ~A~%~
                     so post the following as alternatives.~%~A"
		    var-name other-values)

    ;; Bind var-name to choice and if that fails try the next value.
    ;; First we post an alternative for the values other than choice.
    ;; Then, if we cannot bind to choice, we get rid of that alternative
    ;; and try the next value.
    (let ((pre-change-context (db-request :get-context)))
      (db-request :begin-alt-transaction)
      ;; Post an alt.  This will push a context.
      (with-ag-copy (ag ag)
	(setf (ag-info ag)
	      (list other-values		;values to consider
		    nil))			;not force-auto
	(setf (ag-stage ag) 0)
	(setf (ag-branch-1 ag) (length other-values))
	(post-alternatives ag))
      (if (set-binding var-name choice)
	  ;; Ok, we're outa here, leaving the alt in place.
	  (db-request :commit-alt-transaction)
	(progn
	  ;; The chosen value loses.  Back out and try the next
	  (db-request :abort-alt-transaction)
	  (db-request :pop-context)
	  (assert (eql (db-request :get-context) pre-change-context))
	  (assert (not (db-request :context-was-poisoned)))
	  (dev-warn "We finally win in KS-BIND! -- ~S can't be ~S"
		    var-name choice)
	  (auto-bind ag var-name other-values))))))

(defun set-binding-else-poison (var-name value)
  (MONITOR :KP-1 (:DETAILED "Binding ~A to ~A~%" var-name value))
  (whats-going-on "Binding ~S to ~S" var-name value)
  (unless (set-binding var-name value)
    (dev-warn "Failed to set ~S to ~S." var-name value)
    (post-agenda `(:POISON-STATE :COULD-NOT-FIND-BINDING
		   ,var-name :TO ,value))))

;;; ---------------------------- Change History ----------------------------
;;; (Who)   (When) 		     (What)
;;;
;;;  JWD  10 April 94	Added the code that uses alt-transactions to
;;;			let us try values until one works or we run out.
;;;
