;;;; File: KS-EXPAND.lsp
;;;; SCCS Version: %W%
;;;; Contains: The EXPAND KS.
;;;; Author: Richard Kirby (rbk)
;;;; Created: Fri Jun 22 07:58:42 1990
;;;; Updated: Mon Jul  3 19:57:47 1995 by Jeff Dalton
;;;; Release Version: %Y%
;;;; Copyright: (c) 1992, AIAI, University of Edinburgh
;;;; This material may be reproduced by or for the U.S. Government pursuant
;;;; to the copyright license under the clause at DFARS 252.227-7032
;;;; (June 1975) -- Rights in Technical Data and Computer Software (Foreign).

(in-package :knowledge-source)

(defun get-expand-schemas (pattern)
  (ipc-with-own-copy
    (result (db-request :GET-EXPAND-SCHEMAS pattern)
	    (deep-copy result))))

#+:undef
(defun expand-filters (schema)
  (declare (special ag))
  (let ((result (and (filter-on-only-use-ifs schema
					     (make-ds-node-end
					      :node-tag (cadr (ag-body ag))
					      :node-end :BEGIN))
		     (list schema))))
    (if (null result)
	(dev-debug-msg (format nil "Throwing away ~S"
			       (schema-name (car schema)))
		       *whoami* :information))
    result))

(defun expand-filter (schema bindings)
  (let ((expanding-node (cadr (ag-body *ag*))))
    (filter-on-only-use-ifs
       schema
       bindings
       (make-ds-node-end :node-tag expanding-node
			 :node-end :BEGIN))))

(defun KS-EXPAND (ag &aux (*ag* ag))
  (STAGE-MANAGER
   (schema-list schema not-again-p)
   (node expansion-pattern waiting-schema-list)
   ag
   
   (setq node (cadr (ag-body ag)))
   (setq expansion-pattern (caddr (ag-body ag)))
   
   (when schema-list
     (whats-going-on "Already have a schema list."))
   
   ;; The first stage.
   (unless schema-list
     (setq schema-list (get-expand-schemas expansion-pattern))
     (if (not (listp schema-list)) (setq schema-list (list schema-list))))
   (if (null schema-list)
       (progn
	 (dev-debug-msg (format nil "No schemas found matching ~A~%" node)
			*whoami* :information)
	 (post-agenda '(:POISON-STATE :NO-SCHEMA :INITIALLY))
	 (return-from KS-EXPAND nil)))
   
   (STAGE (if-want-real-time-response-p))
   ;; The second stage.
   ;; Apply filtering rules to each schema.
   ;;   (setq schema-list (mapcan #'expand-filters schema-list))
   (multiple-value-setq (schema-list waiting-schema-list)
     (filter-schemas schema-list #'expand-filter))
   ;; If we have some schemas that must wait for other agenda entries to be
   ;; processed before proceeding, then post alternatives for each such schema
   ;; and a trigger to wait.
   (unless (and (null schema-list) not-again-p)
     ;; Only try this once as long as there is not a change to the
     ;; waiting-schema-list.
     (dolist (schema waiting-schema-list)
       (post-alternatives
	 (make-agenda-rec :id (ag-id ag)
			  :trigger (cdr schema)
			  :body (ag-body ag)
			  ;; Make sure that we don't ask again
			  ;; for a list of schemas.
			  :info (list (list (car schema)) nil :NOT-AGAIN)))))
   (if (null schema-list)
       (progn
	 (dev-debug-msg (format nil "Filtered out all schemas for ~A~%"
				(cadr (ag-body ag))) *whoami* :information)
	 (post-agenda '(:POISON-STATE :NO-SCHEMA :AFTER-FILTERING))
	 (return-from KS-EXPAND nil)))
   
   (STAGE (if-want-real-time-response-p))
   ;; The third stage.
   ;; If more than one schema, pick one, and send the rest to the AM as
   ;; alternatives.
   (when (> (length schema-list) 1)
     ;; Just use the first one, and save the rest.
     (dev-debug :information "More than one applicable schema for ~A~%"
		(cadr (ag-body ag)))
     (whats-going-on "More than one applicable schema for ~A~%~
                      so post alternatives."
		     (cadr (ag-body ag)))
     (setq schema-list (sort-schema-alternatives schema-list))
     (with-ag-copy (ag ag)
       (setf (ag-info ag) (list (cdr schema-list)))
       (setf (ag-branch-1 ag) (length (cdr schema-list)))
       ;; Hack to force alternatives to be recognised.
       (setf (ag-stage ag) 0)
       (post-alternatives ag)))
   (setq schema (car schema-list))
   ;; To save saving it if we should stage.
   (setq schema-list nil)
   
   (STAGE (if-want-real-time-response-p))
   ;; The fourth stage. This is the final stage where the Plan State is
   ;; changed.
   ;; We now have a schema to try. First add in all the nodes, and then see
   ;; if any work is involved in adding the schema's conditions and effects.
   ;; If possible and easily satisfied then continue, otherwise need to do
   ;; some more work but how is the question!
   (setq schema (instantiate-schema schema))
   
   (setq node (cadr (ag-body ag)))
   ;; Expand the schema.
   (whats-going-on "Using ~S as schema for EXPANDING ~S" 
		   (schema-name (car schema)) node)
   (if (expand-from-schema node (car schema))
       (progn
	 (dev-debug-msg
	  (format nil "Expanded ~A OK.~%" node) *whoami* 
	  :information))
       (progn
	 (dev-debug-msg (format nil "Error in expanding ~A~%" node)
			*whoami* :information)
	 (post-agenda '(:POISON-STATE :NO-EXPAND))
	 (return-from KS-EXPAND nil)))
   
   (unless (process-schema (car schema) node t)
     (return-from KS-EXPAND nil))
   
   ;; Now post BINDs for any PSV mentioned.
   (if (listp (cdr schema))
       (dolist (var (cdr schema))
	 (when (psvar-p (var-value var))
	   (post-bind `(:BIND ,(var-value var))))))))

;;;; ---------------------------- Change History ----------------------------
;;;; (Initials) (Date of change)           (Comment)
;;;;    rbk        31/8/92           Updated to include the pattern being
;;;;                                 expanded.
