;;;; File: kp_toplevel.lsp
;;; SCCS Version: %W%
;;; Contains: Top-level processing for the KP.
;;; Author: Richard Kirby (rbk) and Jeff Dalton
;;; Created: Tue Apr 17 15:28:10 1990
;;; Updated: Fri Jul  7 19:08:15 1995 by Jeff Dalton
;;; Release Version: %Y%
;;; Copyright: (c) 1992, AIAI, University of Edinburgh
;;; This material may be reproduced by or for the U.S. Government pursuant
;;; to the copyright license under the clause at DFARS 252.227-7032
;;; (June 1975) -- Rights in Technical Data and Computer Software (Foreign).

(in-package :knowledge-platform)

(import '(ads:make-ds-node-end ads:ds-node-end ads:ds-node-tag))

(use-package '(:ADS :TGM :QA))

(defvar *loaded-ks* nil)

(defun kp-1-startup ()
  (kp-startup))

(defun kp-startup ()
  
  ;; This is the initial function for a KP process.  After startup,
  ;; it is not called again.  Instead, kp-event-handler is called
  ;; whenever a message arrives.

  (unless (is-init *whoami*)
    (dev-error *whoami* "is-init-failed."))
  (setf (component-init-fn *pprocess*) #'kp-init)
  (setf (component-interrupt-fn *pprocess*) #'kp-interrupt-handler)
  ;; Make breaks, etc. go to the debug window.
  (x-redirect-errors :kpout)
  ;; Register with IM.
  (dev-debug-msg "Connecting to IM via MCC." *whoami* :information)
  
  (clear-screen *debug-io*)

  ;; N.B. A KP has it's own value for the current context.
  ;; Set context to the top, and then push to a new one.
  (setq *cucontext* *globalcontext*)
  (pushcontext)

  (mcc-register-with-im)

  (ipc-send :kpready :KPREADY *whoami*)

  (ipc-set-event-handler 'kp-event-handler))


(defun kp-init ()
  ;; Called by KS-INIT for whatever KP runs KS-INIT.  Else called
  ;; my the mcc-client code.
  (clear-screen *debug-io*)
  (initialize-variable-group :KP)
  )


(defun kp-event-handler (self)
  (let* ((e (next-event self))
	 (m (message-contents e)))
    (case (message-route e)
      (:mcc (mcc-handler m))
      (otherwise
       (unwind-protect (catch :ks-exit
			 (handle-kp-message self m))
	 ;; Make sure we send a :KPREADY message if we throw back to
	 ;; the scheduler after an error.
	 (ipc-send :kpready :KPREADY *whoami*))))))

(defun handle-kp-message (self mcc-msg)
  ;; We don't use ipc-handle-mcc-message, because KPs have different
  ;; names (:kp-1, :kp-2, etc); and the IPC message code uses the name
  ;; to find the handler.  /\/
  (declare (ignore self))
  (apply (ipc-get-handler :kp (mcc-msg-ident mcc-msg))
	 (mcc-msg-args mcc-msg)))


(defmessage ((:kp :kpready) :agenda) (entry)

  ;; Looks at the entry to determine the KS required.  If it is not already
  ;; loaded then the $OPLANKSDIR directory is searched for the KS code file.
  ;; If found this is loaded and then called.

  (setq entry (ipc-with-own-copy (ae entry (copy-agenda-rec ae))))
  
  (let ((ks (ensure-loaded-ks (car (ag-body entry)))))
    (MONITOR :KP-1
       (:ONE-LINER "~W/~W,~A: ~W~%" 
		   (ag-id entry) (ag-stage entry) (ag-level entry)
		   (ag-body entry)))
    (whats-going-on "~%**************************************************~%~
                     Processing:~%~S - ~S"
		    (ag-id entry) (ag-body entry))
    (funcall ks entry)))


(defun ensure-loaded-ks (ks-name)
  (assert (keywordp ks-name))
  (unless (member ks-name *loaded-ks*)
    (dev-debug :information "~A not currently loaded.~%" ks-name)
    (push ks-name *loaded-ks*)
    (let* ((ks-filename (concat-string "KS-" ks-name))
	   (filename
	    (concat-string
	     (get-parameter :oplan-ks-dir) "/" ks-filename)))
      (unless (get-parameter :subr)
	(format t "~&Loading ~A~%" ks-filename))
      (if (load-most-recent filename
	     :verbose nil
	     :if-does-not-exist nil)
	  (dev-debug :information "Found KS code file ~A~%" filename)
	(progn
	  (dev-debug :fatal-error "Unable to find KS code file ~A~%" filename)
	  (cerror "Will call ~A."
		  "Please write the KS, and load it in."
		  ks-name)))))
  (values
   (intern (concatenate 'string "KS-" (symbol-name ks-name))
	   (find-package :knowledge-source))))


;;;; Modes

;;; PSV binding mode

(definit :kp *psv-binding-mode* :auto
  "Either :auto or :ask - if :auto then KS-BIND automatically picks a
   variable binding, otherwise the user is presented with the options
   and asked to select a binding.")

;;; Schema selection mode

(definit :kp *schema-selection-mode* :auto)

;;; Poison-handler mode

(definit :kp *poison-handler-mode* :auto)


;;;; Interrupt handler [obsolete /\/]

;;; Tell the AM to add an :USER :USER-REQUEST entry to the agent agenda.

(defun kp-interrupt-handler ()
  (break "Shouldn't get here any more")
  (post-agent-agenda `(:USER :USER-REQUEST)))


#|
(defun agenda-entry-pretty-printer (stream ae)
  (case mon:*monitor-level*
    ((:ONE-LINER :MINIMUM)
     (mon-format stream "~W/~W: ~W"
		 (ag-id ae) (ag-stage ae) (ag-body ae)))
    (t (mon-format stream "~S" ae))))

(eval-when (load eval)
  (mon-add-user-def-print-fn 'agenda-rec #'agenda-entry-pretty-printer))
|#

;;; ---------------------------- Change History ----------------------------
;;; (Who)   (When) 		     (What)
;;;

