;;;; File: main.lsp
;;; SCCS Version: %W%
;;; Contains: O-Plan main program
;;; Author: Jeff Dalton
;;; Created: February 1993
;;; Updated: Sat May  6 21:59:08 1995 by Jeff Dalton
;;; Release Version: %Y%
;;; Copyright: (c) 1992, AIAI, University of Edinburgh
;;; This material may be reproduced by or for the U.S. Government pursuant
;;; to the copyright license under the clause at DFARS 252.227-7032
;;; (June 1975) -- Rights in Technical Data and Computer Software (Foreign).

;;; O-Plan main program

;;; /\/: See also "top-level.lsp".

;;; There are a couple of major variations that can occur:
;;;
;;;  1. We could be running as the Executation Agent (Exec)
;;;     instead of as the Planning Agent (Planner / O-Plan).
;;;
;;;  2. We could be running in connected mode.
;;;
;;; To run O-Plan as the Exec, use the command-line argument "-asexec".
;;; The test used for this case in the O-Plan code is "(is-exec-p)".
;;;
;;; To run O-Plan in conected mode, use the command-line argument "-connect".
;;; The test for this case is "(is-connect-p)".
;;;
;;; In connected mode, standard input and output are used for communicating
;;; with an agent to the left.  In "disconnected" mode, they are used for
;;; a read-evel-print loop ("Lisp listener") process.  By running an agent
;;; in connected mode, it is possible to run it as a Unix process and
;;; communicate with it via pipes.  This is the way the Planner normally
;;; runs the Exec.
;;;

(in-package :oplan)


;;; Component definition

;;; Set-up-components creates the component processes.  Note that some
;;; processes such as the TA that are not, strictly speaking, components
;;; of the planning / exec agent are nonetheless treated as components
;;; here.

;;; In the single-process version, COMPONENT is a subtype of PPROCESS
;;; (pseudo-process), and the order in which pprocesses are created is
;;; also the order in which they are considered for running.

;;; [/\/ There's a hidden dependency in the order chosen below that should
;;; eventually be dealt with in some other way.  The DM is before the AM
;;; so that when the IM sends the DM an :interrupt message, the DM will
;;; handle the message when running in the normal way and not when called
;;; as a subroutine by the AM (which might happen were their order reversed).

;;; This works because the IM is never called as a subroutine.  So if it
;;; was able to send the message, it must have been called directly by
;;; pprocess-main-loop, and the same will be true of the DM -- if it runs
;;; between the IM and the AM.  If the AM gets there first, it might call
;;; the DM as a subroutine, and the DM would see the :interrupt when in
;;; that state instead.

;;; This matters because we want pprocesses to handle :interrupts when
;;; it's easy for them to let other pprocesses run while they're waiting
;;; for user input.  (It should be possible for > 1 pprocess to be in
;;; an interrupt loop at once.)  If a pprocess is running as a subroutine,
;;; nothing else can run until it returns.  So it would have to defer
;;; the point where it enters its interrupt loop by setting a flag,
;;; checking the flag whenever it's running as the top-level pprocess, etc.

;;; Of course, this reasoning would fall apart if, for instance, the
;;; DM ever called the AM as a subroutine (:single-step for the AM is
;;; very like :interrupt for the DM); but it never does.  It may seem
;;; unsafe to rely on this, but calling a pprocess as a subroutine is
;;; so restrictive that it has to be reserved for a small number of
;;; special cases or else the whole structure of the single-process
;;; system would be unworkable. /\/]

(defun set-up-components ()
  ;; Ideally should perhaps read the config file to see what
  ;; components should exist and should allow multiple KPs. /\/

  (unless (or (is-exec-p) (is-connect-p) (is-subr-p))
    (start-component :ta
      :directory "external"
      :debug-level 0))

  (start-component :im
    :directory "interface-manager"
    :debug-level 1)

  (start-component :dm
    :directory "database-manager"
    :debug-level 1)

  (start-component :am
    :directory "agenda-manager"
    :debug-level 1)

  (start-component :kp-1
    :directory "ks-platform"
    :debug-level 1)

  (when (is-get-config-entry :micro-exec)
    (start-component :micro-exec
      :directory "single-process"
      :debug-level 0))

  (when (is-exec-p)
    (start-component :di
      :directory "diary"
      :debug-level 0))

  (when (or (is-exec-p) (is-get-config-entry :micro-exec))
    (if (get-parameter :separate-world)
	(world:start-separate-world)
      (start-component :world
        :directory "world"
	:debug-level 0)))
  )

(defun start-component (name &rest initargs)
  (let ((c (apply #'new-component name initargs)))
    ;; Initialize-component sets up sockets and then assigns a 
    ;; new run-function that performs all subsequent activity.
    (setf (pprocess-run-function c) #'initialize-component
	  (pprocess-status c) :run)
    c))


;;; Component initialization

(defun initialize-component (self)
  ;; /\/: Need a better way of setting these variables.
  (setq oplan::*whoami* (pprocess-name self))
  (setq dev::*dev-debug-level* (component-debug-level self))
  (setq util::*pop-gensymtable* (make-pop-gensymtable))
  ;; We don't need separate component patch directories in the 
  ;; single-process version, and it's simpler and faster not to
  ;; have them.
  ; (load-component-patches self)
  (call-component-startup self))

(defun call-component-startup (self)
  (funcall
    (concat-name (component-name self) '-startup)))

(defun load-component-patches (self)
  ;; N.B. Would load KP patches twice if there were two knowledge platforms.
  (is:is-patch-loader (component-directory self)))


;;; Run

;;; Component is a subclass of pprocess.

;;; Initialization works in a way that depends on the cyclic scheduling
;;; of the pseudo-process mechanism: each pprocess is run once per cycle
;;; and always in the same order (except where one process calls another
;;; as a subroutine).  Consequently, when a pprocess is running it can
;;; get away with assuming that all other runnable pprocesses will run
;;; before it gets to run again.

;;; Components start with status :run so that they can initialize
;;; themselves.  Initialization happens on the first cycle.  All
;;; sockets are created during this cycle, but no events are enqueued.
;;; Initialization also changes the component's run-function and
;;; status.  After the first cycle, all components have status
;;; :run-on-event and a run-function that will handle exactly one
;;; event, then return.  These run-functions no longer depend on
;;; cyclic scheduling.

;;; The TA listens for commands from the user in addition to handling
;;; ipc events.  The pprocess-input-streams mechanism allows this to
;;; be done without (explicit) polling.  (The implementation might
;;; end up polling if a true select function is not available, but
;;; this is not visible at the level of p-processes.)

(defun run ()
  (is:is-configure)			;handle .config file
  (is:is-patch-loader ".")		;load from source/./patch
  (mon-init-db)
  (set-up-components)
  (when (is-connect-p)
    (connect-to-left-io))		;must be after IM (the :leftin server)
  (set-up-repl)
  (run-lights-if-wanted)
  (handler-bind ((error #'top-level-error-handler))
    (pprocess-main-loop)))
  
(defun set-up-repl ()
  (unless (or (is-subr-p) (get-parameter :no-windows))
    (new-repl
      :start-fn 'print-repl-greeting)))

(defun run-lights-if-wanted ()
  (unless (or (get-parameter :no-run-lights)
	      (get-parameter :no-windows))
    (let ((configured-args-if-any (is-get-window-args :run-lights)))
      (when configured-args-if-any
	(run-lights-on configured-args-if-any)))))

;;; An error handler that presents some options to the user

(defun top-level-error-handler (err)
  (when (and *pprocess*
	     (not (typep *pprocess* 'repl))
	     (not (is-subr-p))
	     (get-parameter :interactive))
    (format *debug-io* "~&~%Lisp error: ~A~%~%" err)
    (case (menu-request
	    `("-heading" ,(format nil "Lisp error in process ~S"
			   (pprocess-name *pprocess*))
	      "Allow the error=:allow"
	      "Return to scheduler=:return"
	      "Force global reset=:reset"))
      (:allow
       nil)				;decline
      (:return
       (throw :pprocess-cycle-exit nil))
      (:reset
       (force-reset)))))

;;; Greeting functions

(defun print-repl-greeting (repl-pprocess)
  (terpri (repl-io repl-pprocess))
  (print-oplan-greeting (repl-io repl-pprocess)))

(defun print-oplan-greeting (stream)
  (format stream
	  "~&O-Plan version ~A~%~
             Release date: ~9,,,'0@A~%~
             Build date:   ~9,,,'0@A~%"
	  user::*oplan-version*
	  user::*oplan-release-date*
	  (oplan-load-date)))

(defun get-oplan-load-date () ; -> (date month-name year)
  (declare (notinline simple-defsystem:system-load-date)) ;in case redefined
  (time-util:decode-date-short
    (simple-defsystem:system-load-date
      (simple-defsystem:find-system 'oplan::oplan))))

(defun oplan-load-date ()
  (format nil "~:(~{~A~^-~}~)" (get-oplan-load-date)))

;;; End
