;;;; File: allegro-util.lsp
;;; SCCS Version: %W%
;;; Contains: Lucid-specific utilities
;;; Author: Jeff Dalton <J.Dalton@ed.ac.uk>
;;; Created: April 1994
;;; Updated: Wed Jul 20 16:45:39 1994 by Jeff Dalton
;;; Release Version: %Y%
;;; Copyright: (c) 1994, AIAI, University of Edinburgh
;;; This material may be reproduced by or for the U.S. Government pursuant
;;; to the copyright license under the clause at DFARS 252.227-7032
;;; (June 1975) -- Rights in Technical Data and Computer Software (Foreign).

;;; Some Allegro-specific utilities

;;; A version of this file should exist for each supported CL.

(in-package :util)

;;; INT-CHAR, because X3J13 deleted it and Allegro tries to follow
;;; the proposed standard.  Fortunately, Allegro retained INT-CHAR
;;; in a backwards-compatibility package.

(defun int-char (i)
  (cltl1:int-char i))


;;; Random Unix stuff

(defun argc ()
  (system:command-line-argument-count))

(defun argv (n) ; -> string or nil
  (if (< n (argc))
      (system:command-line-argument n)
    nil))

(defun exit-lisp (&optional (status 0))
  (excl:exit status))

(defun getenv (string) ; -> string or nil
  (system:getenv string))

(defun working-directory-pathname ()
  (excl:current-directory))

(defun change-working-directory (namestring)
  (excl:chdir namestring))


;;; System -- runs a shell command given as a string.

(defun system (cmd)
  (excl:shell cmd))


;;; Interface to xmenu.

;;; It looks like LCL knows to wait (eventaully) for programs run
;;; with :wait nil.  Keeps a record of the child processes, perhaps?

(defvar *xmenu-pathname* nil)

(defun menu-request (xmenu-args &key (read-function #'read))
  (ensure-xmenu-pathname)
  (let ((process-stream
	 (excl:run-shell-command
	    (concat-shell-command *xmenu-pathname*
				  (apply-xmenu-defaults xmenu-args))
	    :output :stream
	    :input nil
	    :error-output nil
	    :wait nil)))
    (unwind-protect
	 (funcall read-function process-stream)
      (close process-stream :abort t))))

(defun apply-xmenu-defaults (xmenu-args)
  (if (or (null (get-parameter :menu-geometry))
	  (member "-geometry" xmenu-args :test #'string=))
      xmenu-args
    (list*
      "-geometry" (get-parameter :menu-geometry)
      xmenu-args)))

(defun ensure-xmenu-pathname ()
  (or *xmenu-pathname*
      (setq *xmenu-pathname*
	    (concatenate 'string (get-parameter :oplan-dir) "/bin/xmenu"))))


;;; General function for getting an i/o stream for communication with
;;; a Unix process.

(defun unix-process-io (program-name &rest args)
  (multiple-value-bind (io-stream error-stream pid)
      (excl:run-shell-command
         (concat-shell-command program-name args)
	 :input :stream
	 :output :stream
	 :error-output nil
	 :wait nil)
    (assert (eq error-stream nil))
    (values
      io-stream
      pid)))


;;; Unlike the Lisps in which O-Plan started out, Allegro CL's
;;; function for running a Unix process expects the entire command as
;;; a single string rather than having one string that's the program
;;; name, plus a list of argument strings.  So we have to take all of
;;; those strings and produce a single string from them.  This is
;;; difficult to get right, since we may need to add quotes or whatever.
;;; So we don't try to get it perfectly right, just to do something
;;; that will probably work for every case we actually try, namely to
;;; put double quotes around all arguments.

(defun concat-shell-command (program-name args)
  (apply #'concatenate 'string
	 "exec "
	 program-name
	 (mapcan #'(lambda (arg)
		     ;; space open-" arg close-"
		     (list " \"" arg "\""))
		 args)))

;;; Structure operations

;;; These functions have to work for "general" structures, where all
;;; slots are of type T.  They needn't work for structures with more
;;; specialized slot types where slots might be stored in unusual ways.

(defun structurep (obj)
  nil)					;/\/ pretend nothing's a structure

(defun map-structure (f s)
  (error "Can't map-structure"))

;;; ---------------------------- Change History ----------------------------
;;; (Who)   (When) 		     (What)
;;;

