;;;; File: context.lsp
;;; SCCS Version: %W%
;;; Contains: A Context Mechanism
;;; Author: Jeff Dalton, based on a version by Ken Currie
;;; Created: Thu Jan 24 14:21:58 1991
;;; Updated: Wed Jul 20 16:45:38 1994 by Jeff Dalton
;;; Release Version: %Y%
;;; Copyright: (c) 1993, AIAI, University of Edinburgh
;;; This material may be reproduced by or for the U.S. Government pursuant
;;; to the copyright license under the clause at DFARS 252.227-7032
;;; (June 1975) -- Rights in Technical Data and Computer Software (Foreign).

;;; *************************************************************************
;;; *                                                                       *
;;; *  CONTEXT.LSP  -  A Context Mechanism                                  *
;;; *                                                                       *
;;; *  This module has its origins in the work on Contexts in POP-2 by      *
;;; *  Harry Barrow.                                                        *
;;; *                                                                       *
;;; *                                                                       *
;;; *************************************************************************

;;; Simple context mechanism, a rational reconstruction of KWC's "obcontext".
;;; The algorithms are essentailly the same, but the code is completely new
;;; and "micro-optimized".  Written by Jeff Dalton.

(in-package :ctxt)

(use-package :util)

(export '(;; Variables
	  *cucontext*			;s.b. *current-context* or *context*
	  *globalcontext*		;s.b. *global-context*

	  ;; Functions
	  newcontext
	  precontext succontexts contextnum
	  contextwithnum		;s.b. find- ?
	  pushcontext popcontext
	  incontext
	  accessincontext updateincontext
	  derefincontext
	  print-context-tree		;was pr_context_tree
	  ; not: killcm prcontext

	  ;; Defining forms
          define-context-accessor

	  ;; Accessors
	  ctxt-symbol-value
	  ctxt-gethash
	  ctxt-puthash
	  ))


;;; Macro to context layer functions

;;; /\/: The field accessor is (fun)called exactly once, to get the
;;; contents of the slot.  It might be better to call it here, in-line,
;;; instead of in accessincontext.  [Ok, let's ... jwd 04 mar 94]

(defmacro define-context-accessor (name field-accessor)
  `(progn
     (proclaim '(inline ,name))
     (defun ,name (o)
       (derefincontext (,field-accessor o)))
     #+:undef
     (defun ,name (o)
       (accessincontext #',field-accessor o))
     (defsetf ,name (o) (new-val)
       `(updateincontext #'(lambda (o v)
			     (setf (,',field-accessor o) v))
			 #',',field-accessor
			 ,o
			 ,new-val))))

;;; Some accessors

(define-context-accessor ctxt-symbol-value symbol-value)

(defun ctxt-gethash (key table &optional default)
  (multiple-value-bind (value found-p) (gethash key table default)
    (values
      (derefincontext value)
      found-p)))

(defun ctxt-puthash (key table value)
  (updateincontext #'(lambda (k v) (setf (gethash k table) v))
                   #'(lambda (k) (gethash k table))
                   key
                   value))

(defsetf ctxt-gethash ctxt-puthash)


;;; Contexts

(defstruct (context (:print-function print-context))
  (number 0 :type fixnum)
  (parent nil)
  (children '()))

#+akcl
(eval-when (compile)
  (si::freeze-defstruct 'context))

(defvar *context-count* 0)

(defvar *globalcontext* (make-context)
  "The initial root context.")

(defvar *root-contexts* (list *globalcontext*)
  "List of contexts with no parent context.")

(defvar *cucontext* *globalcontext*
  "The current context.")


;;; Context printing

(defun print-context (context stream depth)
  (declare (ignore depth))
  (format stream "#<context ~D>" (context-number context)))


;;; Make a new context

(defun newcontext (parent)
  (incf *context-count*)
  (check-type *context-count* fixnum)
  (let ((new-context
	 (make-context :number *context-count*
		       :parent parent
		       :children nil)))
    (if (context-p parent)
	(push new-context (context-children parent))
	(push new-context *root-contexts*))
    new-context))


;;; Public accessors for predecessor, successors, and context number

(defun precontext (c)
  (and (context-p c) (context-parent c)))

(defun succontexts (c)
  (and (context-p c) (context-children c)))

(defun contextnum (c)
  (and (context-p c) (context-number c)))


;;; Return a context with a given number, n.

;;; The search is depth-first, left to right.  However, since a context
;;; always has a higher number than its parent, we do not need to search
;;; the descendents of any context with a number > n.  (The descendents
;;; will also have numbers > n.)  

(defun contextwithnum (n)
  (declare (fixnum n))
  (label c-search ((contexts *root-contexts*))
    (dolist (c contexts)
      (cond ((= (context-number c) n)
	     (return-from contextwithnum c))
	    ((< (context-number c) n)
	     (c-search (context-children c))))))
  (error "No context numbered ~D." n))


;;; Context push and pop

(defun pushcontext ()
  (setq *cucontext* (newcontext *cucontext*)))

(defun popcontext ()
  (when (and (context-p *cucontext*)
	     (not (null (context-parent *cucontext*))))
    (setq *cucontext* (context-parent *cucontext*))))


;;; Apply in context

(defun incontext (c fn &rest args)
  #+akcl (declare (:dynamic-extent args))
  (let ((*cucontext* c))
    (apply fn args)))


;;; Access in context

;;; /\/: A cval could contain a 1-element cache for the value most
;;; recently looked-up.  This would win if lookups significantly
;;; outnumbered updates and context switches.  N.B.  It would cache
;;; the current context (and the value), not the context that
;;; established the value.  [Let's do this too -- jwd 04 mar 94]

(defstruct cval
  "Context value.  Contains a sorted alist indexed by context.  The
   alist is never null.  Newer (higher numbered) contexts appear first."
  alist
  cached-context			;most recently seen context
  cached-value)				;and the associated value

#+akcl
(eval-when (compile)
  (si::freeze-defstruct 'cval))

;;; We want the value for the current context or the nearest ancestor
;;; for which a value's been stored.

;;; Newer contexts have higher numbers.  A context is always newer
;;; than its parent.  In the cval alist, entries for newer contexts
;;; appear first.

;;; /\/: It would be faster if we didn't have to call context-p.
;;; Can we assume it's either a context or NIL?

(defun accessincontext (accessor object)
  (declare (optimize (speed 3) (safety 0) (space 0)))
  (let ((v (funcall accessor object)))
    (if (not (cval-p v))
	;; A value that isn't a cval doesn't depend on context.
	v
      (if (eq (cval-cached-context v) *cucontext*)
	  (cval-cached-value v)
	(let ((target *cucontext*)
	      (alist (cval-alist v))
	      (n 0))
	  (declare (fixnum n))
	  (setf (cval-cached-context v) *cucontext*)
	  (do () ((not (context-p target)) :undef)
	    (setq n (context-number target))
	    ;; Skip enties for contexts newer than the target.
	    (while (and (consp alist) (< n (context-number (caar alist))))
	      (setq alist (cdr alist)))
	    (when (null alist)
	      (return (setf (cval-cached-value v) :undef)))
	    (when (eq target (caar alist))
	      (return (setf (cval-cached-value v) (cdar alist))))
	    ;; All remaining entries are older than the target
	    ;; but perhaps not older than the target's parent.
	    (setq target (context-parent target))))))))

(defun derefincontext (v)
  (declare (optimize (speed 3) (safety 0) (space 0)))
  (cond ((not (cval-p v))
	 ;; A value that isn't a cval doesn't depend on context.
	 v)
	((eq (cval-cached-context v) *cucontext*)
	 (cval-cached-value v))
	(t
	 (let ((target *cucontext*)
	       (alist (cval-alist v))
	       (n 0))
	   (declare (fixnum n))
	   (setf (cval-cached-context v) *cucontext*)
	   (do () ((not (context-p target)) :undef)
	     (setq n (context-number target))
	     ;; Skip enties for contexts newer than the target.
	     (while (and (consp alist) (< n (context-number (caar alist))))
	       (setq alist (cdr alist)))
	     (when (null alist)
	       (return (setf (cval-cached-value v) :undef)))
	     (when (eq target (caar alist))
	       (return (setf (cval-cached-value v) (cdar alist))))
	     ;; All remaining entries are older than the target
	     ;; but perhaps not older than the target's parent.
	     (setq target (context-parent target)))))))


;;; Update in context

(defun updateincontext (updater accessor object new-value)
  (let ((v (funcall accessor object)))
    ;; If the current value, v, is not already context-dependent, then
    ;; make it context-dependent by making it the value for some context.
    ;; If v isn't :undef, save it as the value in the global context.
    ;; Otherwise, just make it the value in the current context, and it
    ;; will be replaced by new-value below.
    (when (not (cval-p v))
      (let ((ctxt (if (eq v :undef) *cucontext* *globalcontext*)))
	(setq v (make-cval :alist (list (cons ctxt v))))
	(funcall updater object v)))
    ;; Cache context and value
    (setf (cval-cached-context v) *cucontext*
	  (cval-cached-value v) new-value)
    ;; Install the value in the alist, adding a new alist entry if necessary
    (let ((alist (cval-alist v))
	  (target *cucontext*)
	  (n (context-number *cucontext*)))
      (declare (fixnum n))
      (assert (not (null alist)))
      (loop
	(cond ((eq (caar alist) target)
	       (setf (cdar alist) new-value)
	       (return new-value))
	      ((< (context-number (caar alist)) n)
	       (setf (cdr alist) (cons (car alist) (cdr alist)))
	       (setf (car alist) (cons target new-value))
	       (return new-value))
	      ((null (cdr alist))
	       (setf (cdr alist) (list (cons target new-value)))
	       (return new-value))
	      (t (setq alist (cdr alist))))))))


;;; Print context tree

;;; May want parameters that specify a root and a depth limit.

(defun print-context-tree ()
  (label walk ((contexts *root-contexts*)
	       (level 0))
    (dolist (c (reverse contexts))
      (format t "~&~vT~S~%" (* 3 level) (contextnum c))
      (walk (context-children c)
	    (+ level 1)))))


;;;; HISTORY:
;;
;;  2-Oct-89:  Converted to packages by KWC.
;;  5-Oct-89:  Complete re-write to be compatible with the POP package
;;             used in O-Plan1.
;;  7-Nov-89:  Included a Context Tree Printer, after much debugging.
;;    1991  :  Complete rewrite by JWD who wanted to understand how it
;;             worked and to try various micro-optimizations.
;;   May 93 :  New ('91) version adapted to O-Plan for release 2.0.
;;   Oct 93 :  Minor cleanups.
;;  4 Mar 94:  Added derefincontext and the one-context cache.
;;  8 Mar 94:  Added ctxt-symbol-value.
;; 26 May 94:  Added ctxt-gethash.
;;
;;;;

;;; End

