;;;; File: kcl-xwindowio.lsp
;;; SCCS Version: %W%
;;; Contains: Code for using X windows for IO.
;;; Author: Richard Kirby (rbk) and Jeff Dalton
;;; Created: Fri Jan 12 17:25:00 1990
;;; Updated: Sat May  6 23:36:11 1995 by Jeff Dalton
;;; Release Version: %Y%
;;; Copyright: (c) 1992, 1993, AIAI, University of Edinburgh
;;; This material may be reproduced by or for the U.S. Government pursuant
;;; to the copyright license under the clause at DFARS 252.227-7032
;;; (June 1975) -- Rights in Technical Data and Computer Software (Foreign).

(in-package :util)

;;; *xterm-debug* is a stream for debugging output produced when
;;; trying to start xterms.  It's normally bound to a stream that
;;; discards all output, obtained by calling make-broadcast-stream
;;; with no arguments.

(defvar *xterm-debug* (make-broadcast-stream))

;; A readmacro for Clines.
(eval-when (compile)
  (set-macro-character #\%
     #'(lambda (stream char)
	 (declare (ignore char))
	 (values (read-line stream)))))

(Clines
 % static int fd_from_filedescr(lispstream)
 %   object lispstream;
 % {
 %   return(fileno(lispstream->sm.sm_fp));
 % }
)

#-|FreeBSD|
(Clines
 %#include <sys/types.h>
 %#include <sys/stat.h>
 %#include <sys/file.h>
 %#ifdef __386BSD__
 %#define USE_OLD_TTY
 %#endif
 %#include <sys/ioctl.h>
 %#include <fcntl.h>
 %enum smmode {			/*  stream mode  */
 %	smm_input,		/*  input  */
 %	smm_output,		/*  output  */
 %	smm_io,			/*  input-output  */
 %	smm_probe,		/*  probe  */
 %	smm_synonym,		/*  synonym  */
 %	smm_broadcast,		/*  broadcast  */
 %	smm_concatenated,	/*  concatenated  */
 %	smm_two_way,		/*  two way  */
 %	smm_echo,		/*  echo  */
 %	smm_string_input,	/*  string input  */
 %	smm_string_output	/*  string output  */
 %};
 
 % static int set_term_modes(lispstream)
 %   object lispstream;
 % {
 %   int err;
 %   struct sgttyb sb;
 %   struct tchars tc;
 %   struct ltchars ltc;
 %   int bits, from, to;
 
 %   switch (lispstream->sm.sm_mode)
 %      {
 %      case smm_two_way:
 %         to = fileno(lispstream->sm.sm_object0->sm.sm_fp);
 %         break;
 %      default:   
 %         to = fileno(lispstream->sm.sm_fp);
 %      }
 %   from = open("/dev/tty", O_RDWR, 0);
 %   if (from == -1) {
 %      err = -1;
 %      goto error;
 %   }
 %   err = ioctl(from, TIOCGETP, &sb);
 %   err |= ioctl(to, TIOCSETP, &sb);
 %   err |= ioctl(from, TIOCGETC, &tc);
 %   err |= ioctl(to, TIOCSETC, &tc);
 %   err |= ioctl(from, TIOCLGET, &bits);
 %   err |= ioctl(to, TIOCLSET, &bits);
 %   err |= ioctl(from, TIOCGLTC, &ltc);
 %   err |= ioctl(to, TIOCSLTC, &ltc);
 %error:
 %   if (err != 0) perror("Oh dear, failed to set terminal modes.\n");
 %   close(from);
 %   return err;
 % }
)

(defentry fd-from-filedescr (object)
  (int fd_from_filedescr))

#-|FreeBSD|
(defentry set-term-modes (object)
  (int set_term_modes))

#+|FreeBSD|
(defun set-term-modes (tty-name)
  (system (concat-string "stty -f " tty-name " sane")))


;;; Changes for 1.2 release:
;;;  * More error checks.
;;;  * Moved close of pseudo-term
;;;  * Uses FORK:PROCESS instead of SYSTEM.
;;;  * PROGified it so we can do (go FAIL) and (go RETRY).
;;;  * Waits for xterm.  For some reason, the READ was getting an EOF
;;;    instead of waiting until some output was produced, perhaps because
;;;    the xterm was exiting.  We now wait until input is available
;;;    (according to LISTEN) before calling READ and also check for
;;;    failed xterms.

;;; Note that a RETRY may use the same pseudo-terminal that was used
;;; before.  This seems to work (so far).  This suggests that we might be
;;; able to just rerun xterm without looking for a new pty.

;;; The following shell command is useful for seeing what's happening
;;; with xterm.  Substitite $OPLANTMPDIR for oplan-tmp.
;;;   gnugrep xterm oplan-tmp/*.out oplan-tmp/exec/*.out

;;; When debugging, note that you can run /aiai/bin/X.V11R5/xterm
;;; via "trace", and that "trace" can easily be exchanged for "env"
;;; as a quick way to turn off the trace output.  Here's what it
;;; would look like:

#|
    ;; Run xterm
    (setq xterm-pid
	  (car
	    (fork:process "env"		;or use "trace"
	      :args `("/aiai/bin/X.V11R5/xterm"
		      ,(concatenate 'string "-S"
			pseudo-term-letter
			pseudo-term-number
			(prin1-to-string (fd-from-filedescr pseudo-term)))
		      ,@(if (listp args) args (fork:break-args args))))))
|#

(defun run-xterm-for-io (&optional (args ""))
  "Opens a window for input and output. 
   Arguments:
     args - A string or list containing extra args for xterm.
   Returns:
     A bidirectional stream, and the device name of the tty device."
  
  (declare (type (or string list) args))
  
  (prog (pseudo-term io-term pseudo-term-name io-term-name
         pseudo-term-letter pseudo-term-number status xterm-pid)
    ;; Find a pseudo terminal to use.
    RETRY
    (block found-a-term
      (dolist (letter '("p" "q" "r"))
	(dolist (number '("0" "1" "2" "3" "4" "5" "6" "7" "8" "9" 
			  "a" "b" "c" "d" "e" "f"))
	  (if (and (setq pseudo-term
			 (open (setq pseudo-term-name
				     (concatenate 'string
						  "/dev/pty"
						  letter 
						  number))
			       :direction :io
			       :if-exists :append
			       :if-does-not-exist nil))
		   (progn
		     (setq pseudo-term-letter letter)
		     (setq pseudo-term-number number)
		     (setq io-term-name
			   (concatenate 'string "/dev/tty"
					letter number))
		     (setq io-term
			   (open io-term-name :direction :io
				 :if-exists :append
				 :if-does-not-exist nil))))
	      (return-from found-a-term))))
      (error "Can't find a free pseudo terminal."))
    #-|FreeBSD|
    (set-term-modes io-term)		;<<< New
    #+|FreeBSD|
    (set-term-modes io-term-name)
    (close io-term)

    (format *xterm-debug* "~&Using /dev/pty~A~A.  "
	    pseudo-term-letter pseudo-term-number)

    ;; Run xterm
    (setq xterm-pid
	  (car
	    (fork:process "xterm"
	      :args `(,(concatenate 'string "-S"
			pseudo-term-letter
			pseudo-term-number
			(prin1-to-string (fd-from-filedescr pseudo-term)))
		      ,@(if (listp args) args (fork:break-args args))))))
    
    ;; (close pseudo-term)

    ;; For some reason the following doesn't work on the Suns.
    ;; Listen always returns true and read seems to read the output
    ;; that the planner sent (chiefly messages about how you should
    ;; type 1 or 8).  /\/
    ;; (setq io-term
    ;;       (open io-term-name :direction :io :if-exists :append))

    (setq io-term
	  (make-two-way-stream
	     (open io-term-name :direction :input)
	     (open io-term-name :direction :output :if-exists :append)))
    
    ;; Wait for some gumpf from xterm, so that we know that it has started ok.
    (let ((failures 0)
	  input)
      (loop
        (when (listen io-term)
	  (setq input (read io-term nil :xeof))
	  (unless (eq input :xeof)
	    (return)))
	(incf failures)
	(if (= failures 1)
	    (format *xterm-debug* "Waiting for xterm. ")
	  (format *xterm-debug* "."))
	(sleep 1)
	(setq status (fork:waitpid-no-hang xterm-pid))
	(when (= (car status) xterm-pid)
	  (go FAIL))))
    ;; Success!
    (format *xterm-debug* " Have xterm; pid = ~S.~%" xterm-pid)
    (register-xterm-pid xterm-pid)
    (close pseudo-term)			;<<< Moved from above
    #-|FreeBSD|
    (set-term-modes io-term)
    #+|FreeBSD|
    (set-term-modes io-term-name)
    (return
      (values io-term io-term-name))
    FAIL
    (format *xterm-debug* "~&xterm failed.~%")
    (format *xterm-debug* "~&xterm status: ~S.~%" status)
    (close pseudo-term)
    (close io-term)
    (go RETRY)))


;;; Make break-args from support/xfork available as a util function.

(defun break-args (string)
  (fork:break-args string))


;;; Child process operations

;;; Record of xterm pids

(defvar *xterm-pids* '())

(defun register-xterm-pid (pid)
  (assert (not (member pid *xterm-pids*)))
  (setq *xterm-pids* (nconc *xterm-pids* (list pid))))

;;; The xterms exit if they get an INT signal.  HUP isn't enough on Suns.

(defun terminate-xterms-if-necessary ()
  (cond ((string= (getenv "HOSTTYPE") "i386")
	 (let ((live-pids (remove-if-not #'still-running-p *xterm-pids*)))
	   (when live-pids
	     (system
	       (format nil "kill -INT ~{~D~^ ~}" live-pids)))))
	(t ;; Else (on Suns) don't need to.
	   nil)))

(defun still-running-p (pid)
  (let ((status (fork:waitpid-no-hang pid)))
    ;; /\/: Maybe check (= (car status) 0) instead.
    (not (= (car status) pid))))

;;;; ---------------------------- Change History ----------------------------
;;;; (Initials) (Date of change)           (Comment)
